import string
import os
import json
import numpy as np
import shutil
from ._server import serve

try:
    from html import escape
except ImportError:
    from cgi import escape

_color_sentiment = {
    'positive': 'rgb(143, 255, 113)',
    'neutral': 'rgb(255, 238, 109)',
    'negative': 'rgb(255, 139, 118)',
}
_color_emotion = {
    'anger': 'rgb(254, 0, 20)',
    'fear': 'rgb(168, 103, 172)',
    'joy': 'rgb(0, 211, 239)',
    'love': 'rgb(255, 70, 198)',
    'sadness': 'rgb(255, 206, 0)',
    'surprise': 'rgb(255, 244, 0)',
    'neutral': 'rgb(255, 255, 255)',
}
_color_toxic = {
    'toxic': 'rgb(0, 217, 0)',
    'severe_toxic': 'rgb(244, 248, 0)',
    'obscene': 'rgb(248, 34, 13)',
    'threat': 'rgb(0, 238, 241)',
    'insult': 'rgb(248, 95, 34)',
    'identity_hate': 'rgb(230, 183, 0)',
    'neutral': 'rgb(255, 255, 255)',
}


def _sentiment_mark(text, negative, positive, neutral, attention, label):
    return (
        "<mark style='background-color:%s' class='tooltipped' data-position='bottom' data-tooltip=\"Positive <i class='em em-smiley_cat'></i> %.3f<br>Neutral <i class='em em-cat'></i> %.3f<br>Negative <i class='em em-pouting_cat'></i> %.3f<br>Attention <i class='em em-warning'></i> %.3f\">%s</mark>"
        % (
            _color_sentiment[label],
            positive,
            neutral,
            negative,
            attention,
            text,
        )
    )


def _relevancy_mark(text, negative, positive, attention, label):
    return (
        "<mark style='background-color:%s' class='tooltipped' data-position='bottom' data-tooltip=\"Relevant <i class='em em-information_source'></i> %.3f<br>Not relevant <i class='em em-lying_face'></i> %.3f<br>Attention <i class='em em-warning'></i> %.3f\">%s</mark>"
        % (_color_sentiment[label], positive, negative, attention, text)
    )


def _toxic_mark(
    text,
    toxic,
    severe_toxic,
    obscene,
    threat,
    insult,
    identity_hate,
    attention,
    label,
):
    return (
        "<mark style='background-color:%s' class='tooltipped' data-position='bottom' data-tooltip=\"Toxic <i class='em em-face_vomiting'></i> %.3f<br>Severe Toxic <i class='em em-face_vomiting'></i><i class='em em-face_vomiting'></i> %.3f<br>Obscene <i class='em em-confounded'></i> %.3f<br>Threat <i class='em em-fearful'></i> %.3f<br>Insult <i class='em em-boar'></i> %.3f<br>Identity hate <i class='em em-cry'></i> %.3f<br>Attention <i class='em em-warning'></i> %.3f<br>\">%s</mark>"
        % (
            _color_toxic[label],
            toxic,
            severe_toxic,
            obscene,
            threat,
            insult,
            identity_hate,
            attention,
            text,
        )
    )


def _emotion_mark(
    text, anger, fear, joy, love, sadness, surprise, attention, label
):
    return (
        "<mark style='background-color:%s' class='tooltipped' data-position='bottom' data-tooltip=\"Anger <i class='em em-angry'></i> %.3f<br>Fear <i class='em em-fearful'></i> %.3f<br>Joy <i class='em em-joy'></i> %.3f<br>Love <i class='em em-heart_eyes'></i> %.3f<br>Sadness <i class='em em-white_frowning_face'></i> %.3f<br>Surprise <i class='em em-dizzy_face'></i> %.3f<br>Attention <i class='em em-warning'></i> %.3f<br>\">%s</mark>"
        % (
            _color_emotion[label],
            anger,
            fear,
            joy,
            love,
            sadness,
            surprise,
            attention,
            text,
        )
    )


def _render_binary(data, notebook_mode = False):
    index_negative = data['barplot']['x'].index('negative')
    index_positive = data['barplot']['x'].index('positive')
    index_neutral = data['barplot']['x'].index('neutral')
    sentiment_mark = []
    for k, v in data['word'].items():
        sentiment_mark.append(
            _sentiment_mark(
                k,
                v[index_negative],
                v[index_positive],
                v[index_neutral],
                data['alphas'][k],
                data['barplot']['x'][np.argmax(v)],
            )
        )
    sentiment_mark = ' '.join(sentiment_mark)
    this_dir = os.path.dirname(__file__)

    if notebook_mode:
        js_location, css_location = _upload_jupyter()
    else:
        js_location = 'static/echarts.min.js'
        css_location = 'static/admin-materialize.min.css'

    with open(os.path.join(this_dir, 'web', 'index.html')) as _file:
        template = string.Template(_file.read())

    template = template.substitute(
        label = escape(data['class_name']),
        p = sentiment_mark,
        barplot_positive = escape(
            json.dumps(int(data['barplot']['y'][index_positive]))
        ),
        barplot_neutral = escape(
            json.dumps(int(data['barplot']['y'][index_neutral]))
        ),
        barplot_negative = escape(
            json.dumps(int(data['barplot']['y'][index_negative]))
        ),
        histogram_x = escape(json.dumps(data['histogram']['x'].tolist())),
        histogram_y = escape(json.dumps(data['histogram']['y'].tolist())),
        attention_x = escape(json.dumps(data['attention']['x'].tolist())),
        attention_y = escape(json.dumps(data['attention']['y'].tolist())),
        css_location = css_location,
        js_location = js_location,
    )
    if notebook_mode:
        from IPython.display import display, HTML

        display(HTML(template))
    else:
        serve(template)


def _render_relevancy(data, notebook_mode = False):
    index_negative = data['barplot']['x'].index('negative')
    index_positive = data['barplot']['x'].index('positive')
    relevancy_mark = []
    for k, v in data['word'].items():
        relevancy_mark.append(
            _relevancy_mark(
                k,
                v[index_negative],
                v[index_positive],
                data['alphas'][k],
                data['barplot']['x'][np.argmax(v)],
            )
        )
    relevancy_mark = ' '.join(relevancy_mark)
    this_dir = os.path.dirname(__file__)

    if notebook_mode:
        js_location, css_location = _upload_jupyter()
    else:
        js_location = 'static/echarts.min.js'
        css_location = 'static/admin-materialize.min.css'

    with open(os.path.join(this_dir, 'web', 'index_relevancy.html')) as _file:
        template = string.Template(_file.read())

    template = template.substitute(
        label = escape(data['class_name']),
        p = relevancy_mark,
        barplot_positive = escape(
            json.dumps(int(data['barplot']['y'][index_positive]))
        ),
        barplot_negative = escape(
            json.dumps(int(data['barplot']['y'][index_negative]))
        ),
        histogram_x = escape(json.dumps(data['histogram']['x'].tolist())),
        histogram_y = escape(json.dumps(data['histogram']['y'].tolist())),
        attention_x = escape(json.dumps(data['attention']['x'].tolist())),
        attention_y = escape(json.dumps(data['attention']['y'].tolist())),
        css_location = css_location,
        js_location = js_location,
    )
    if notebook_mode:
        from IPython.display import display, HTML

        display(HTML(template))
    else:
        serve(template)


def _render_toxic(data, notebook_mode = False):
    index_toxic = data['barplot']['x'].index('toxic')
    index_severe_toxic = data['barplot']['x'].index('severe_toxic')
    index_obscene = data['barplot']['x'].index('obscene')
    index_threat = data['barplot']['x'].index('threat')
    index_insult = data['barplot']['x'].index('insult')
    index_identity_hate = data['barplot']['x'].index('identity_hate')
    toxic_mark = []
    for k, v in data['word'].items():
        where = np.where(np.array(v) >= 0.5)[0].shape[0]
        if where:
            where = data['barplot']['x'][np.argmax(v)]
        else:
            where = 'neutral'
        toxic_mark.append(
            _toxic_mark(
                k,
                v[index_toxic],
                v[index_severe_toxic],
                v[index_obscene],
                v[index_threat],
                v[index_insult],
                v[index_identity_hate],
                data['alphas'][k],
                where,
            )
        )
    toxic_mark = ' '.join(toxic_mark)
    this_dir = os.path.dirname(__file__)

    if notebook_mode:
        js_location, css_location = _upload_jupyter()
    else:
        js_location = 'static/echarts.min.js'
        css_location = 'static/admin-materialize.min.css'

    with open(os.path.join(this_dir, 'web', 'index_toxic.html')) as _file:
        template = string.Template(_file.read())

    template = template.substitute(
        label = escape(data['class_name']),
        p = toxic_mark,
        barplot_toxic = escape(
            json.dumps(int(data['barplot']['y'][index_toxic]))
        ),
        barplot_severe_toxic = escape(
            json.dumps(int(data['barplot']['y'][index_severe_toxic]))
        ),
        barplot_obscene = escape(
            json.dumps(int(data['barplot']['y'][index_obscene]))
        ),
        barplot_threat = escape(
            json.dumps(int(data['barplot']['y'][index_threat]))
        ),
        barplot_insult = escape(
            json.dumps(int(data['barplot']['y'][index_insult]))
        ),
        barplot_identity_hate = escape(
            json.dumps(int(data['barplot']['y'][index_identity_hate]))
        ),
        histogram_x = escape(json.dumps(data['histogram']['x'].tolist())),
        histogram_y = escape(json.dumps(data['histogram']['y'].tolist())),
        attention_x = escape(json.dumps(data['attention']['x'].tolist())),
        attention_y = escape(json.dumps(data['attention']['y'].tolist())),
        css_location = css_location,
        js_location = js_location,
    )
    if notebook_mode:
        from IPython.display import display, HTML

        display(HTML(template))
    else:
        serve(template)


def _render_emotion(data, notebook_mode = False):
    index_anger = data['barplot']['x'].index('anger')
    index_fear = data['barplot']['x'].index('fear')
    index_joy = data['barplot']['x'].index('joy')
    index_love = data['barplot']['x'].index('love')
    index_sadness = data['barplot']['x'].index('sadness')
    index_surprise = data['barplot']['x'].index('surprise')
    emotion_mark = []
    for k, v in data['word'].items():
        where = np.where(np.array(v) >= 0.3)[0].shape[0]
        if where:
            where = data['barplot']['x'][np.argmax(v)]
        else:
            where = 'neutral'
        emotion_mark.append(
            _emotion_mark(
                k,
                v[index_anger],
                v[index_fear],
                v[index_joy],
                v[index_love],
                v[index_sadness],
                v[index_surprise],
                data['alphas'][k],
                where,
            )
        )
    emotion_mark = ' '.join(emotion_mark)
    this_dir = os.path.dirname(__file__)

    if notebook_mode:
        js_location, css_location = _upload_jupyter()
    else:
        js_location = 'static/echarts.min.js'
        css_location = 'static/admin-materialize.min.css'

    with open(os.path.join(this_dir, 'web', 'index_emotion.html')) as _file:
        template = string.Template(_file.read())

    template = template.substitute(
        label = escape(data['class_name']),
        p = emotion_mark,
        barplot_anger = escape(
            json.dumps(int(data['barplot']['y'][index_anger]))
        ),
        barplot_fear = escape(
            json.dumps(int(data['barplot']['y'][index_fear]))
        ),
        barplot_joy = escape(json.dumps(int(data['barplot']['y'][index_joy]))),
        barplot_love = escape(
            json.dumps(int(data['barplot']['y'][index_love]))
        ),
        barplot_sadness = escape(
            json.dumps(int(data['barplot']['y'][index_sadness]))
        ),
        barplot_surprise = escape(
            json.dumps(int(data['barplot']['y'][index_surprise]))
        ),
        histogram_x = escape(json.dumps(data['histogram']['x'].tolist())),
        histogram_y = escape(json.dumps(data['histogram']['y'].tolist())),
        attention_x = escape(json.dumps(data['attention']['x'].tolist())),
        attention_y = escape(json.dumps(data['attention']['y'].tolist())),
        css_location = css_location,
        js_location = js_location,
    )
    if notebook_mode:
        from IPython.display import display, HTML

        display(HTML(template))
    else:
        serve(template)


def _attention(attn_data):
    from IPython.core.display import display, HTML, Javascript

    vis_html = """
          <span style="user-select:none">
            Layer: <select id="layer"></select>
          </span>
          <div id='vis'></div>
        """

    display(HTML(vis_html))
    this_dir = os.path.dirname(__file__)
    vis_js = open(
        os.path.join(this_dir, 'web', 'static', 'head_view.js')
    ).read()
    params = {'attention': attn_data, 'default_filter': 'all'}
    display(Javascript('window.params = %s' % json.dumps(params)))
    display(Javascript(vis_js))


def _upload_jupyter():
    location = os.getcwd()

    this_dir = os.path.dirname(__file__)

    js_location = os.path.join(this_dir, 'web', 'static', 'echarts.min.js')
    css_location = os.path.join(
        this_dir, 'web', 'static', 'admin-materialize.min.css'
    )

    shutil.copyfile(js_location, './echarts.min.js')
    shutil.copyfile(css_location, './admin-materialize.min.css')

    return 'echarts.min.js', 'admin-materialize.min.css'



from __future__ import print_function
import errno
import logging
import os
import pickle
import random
import re
import time
import requests
import threading
import json
import urllib.request
from bs4 import BeautifulSoup
from fake_useragent import UserAgent
from newspaper import Article
from datetime import datetime, timedelta
from dateutil import parser
from queue import Queue
from urllib.parse import quote
from unidecode import unidecode

xgb_language = None

NUMBER_OF_CALLS_TO_GOOGLE_NEWS_ENDPOINT = 0

GOOGLE_NEWS_URL = 'https://www.google.com.my/search?q={}&source=lnt&tbs=cdr%3A1%2Ccd_min%3A{}%2Ccd_max%3A{}&tbm=nws&start={}'

logging.basicConfig(
    level = logging.DEBUG, format = '%(asctime)s - %(levelname)s - %(message)s'
)


def get_date(load):
    try:
        date = re.findall(
            '[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?', load
        )
        return '%s-%s-%s' % (date[2], date[0], date[1])
    except Exce:
        return False


def run_parallel_in_threads(target, args_list):
    globalparas = []
    result = Queue()

    def task_wrapper(*args):
        result.put(target(*args))

    threads = [
        threading.Thread(target = task_wrapper, args = args)
        for args in args_list
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    while not result.empty():
        globalparas.append(result.get())
    globalparas = list(filter(None, globalparas))
    return globalparas


def forge_url(q, start, year_start, year_end):
    global NUMBER_OF_CALLS_TO_GOOGLE_NEWS_ENDPOINT
    NUMBER_OF_CALLS_TO_GOOGLE_NEWS_ENDPOINT += 1
    return GOOGLE_NEWS_URL.format(
        q.replace(' ', '+'), str(year_start), str(year_end), start
    )


def extract_links(content):
    soup = BeautifulSoup(content, 'html.parser')
    today = datetime.now().strftime('%m/%d/%Y')
    links_list = [
        v.attrs['href'] for v in soup.find_all('a', {'class': ['lLrAF']})
    ]
    dates_list = [v.text for v in soup.find_all('div', {'class': ['slp']})]
    output = []
    for (link, date) in zip(links_list, dates_list):
        try:
            date = date.split('-')
            if (
                date[1].find('hour') >= 0
                or date[1].find('minute') >= 0
                or date[1].find('分鐘') >= 0
                or date[1].find('小時') >= 0
            ):
                date[1] = today
            elif date[1].find('day') >= 0 or date[1].find('日') >= 0:
                count = date[1].split(' ')[0]
            else:
                try:
                    date[1] = parser.parse(date[1]).strftime('%m-%d-%Y')
                except:
                    date[1] = 'null'
            output.append((link, date[0].strip(), date[1]))
        except:
            continue
    return output


def get_malaya_summary(text):
    import malaya

    return malaya.summarize_lsa(text, important_words = 20)


def get_article(link, news, date):
    article = Article(link)
    article.download()
    article.parse()
    article.nlp()
    lang = 'ENGLISH'
    if len(article.title) < 5 or len(article.text) < 5:
        lang = 'INDONESIA'
        print('found BM/ID article')
        article = Article(link, language = 'id')
        article.download()
        article.parse()
        if xgb_language:
            lang = xgb_language.predict(article.text)
            malaya_summarized = get_malaya_summary(article.text.split('\n'))
            article.summary = malaya_summarized['summary']
            article.keywords = malaya_summarized['cluster-top-words']
        else:
            article.nlp()
    return {
        'title': article.title,
        'url': link,
        'authors': article.authors,
        'top-image': article.top_image,
        'text': article.text,
        'keyword': article.keywords,
        'summary': article.summary,
        'news': news,
        'date': date,
        'language': lang,
    }


def google_news_run(
    keyword,
    limit = 10,
    year_start = 2010,
    year_end = 2011,
    debug = True,
    sleep_time_every_ten_articles = 0,
    xgb_model = None,
):
    global xgb_language
    if xgb_model:
        xgb_language = xgb_model
    num_articles_index = 0
    ua = UserAgent()
    results = []
    while num_articles_index < limit:
        url = forge_url(keyword, num_articles_index, year_start, year_end)
        if debug:
            logging.debug('For Google -> {}'.format(url))
            logging.debug(
                'Total number of calls to Google = {}'.format(
                    NUMBER_OF_CALLS_TO_GOOGLE_NEWS_ENDPOINT
                )
            )
        headers = {'User-Agent': ua.chrome}
        success = False
        try:
            response = requests.get(url, headers = headers, timeout = 60)
            if (
                str(response.content).find(
                    'In the meantime, solving the above CAPTCHA will let you continue to use our services'
                )
                >= 0
            ):
                print('whops, blocked')
                return results
            links = extract_links(response.content)
            nb_links = len(links)
            if nb_links == 0 and num_articles_index == 0:
                print(
                    'No results fetched. Either the keyword is wrong or you have been banned from Google. Retry tomorrow or change of IP Address.'
                )
                return results
            if nb_links == 0:
                print('No more news to read for keyword {}.'.format(keyword))
                return results
            for link in links:
                try:
                    results.append(get_article(*link))
                except:
                    pass
            success = True
        except requests.exceptions.Timeout:
            logging.debug(
                'Google news Timeout. Maybe the connection is too slow. Skipping.'
            )
            continue
        num_articles_index += 10
        if debug and sleep_time_every_ten_articles != 0:
            logging.debug(
                'Program is going to sleep for {} seconds.'.format(
                    sleep_time_every_ten_articles
                )
            )
        time.sleep(sleep_time_every_ten_articles)
    return results



import json
from ._utils import (
    check_file,
    load_graph,
    check_available,
    generate_session,
    sentencepiece_tokenizer_bert,
    sentencepiece_tokenizer_xlnet,
)
from .._models._bert_model import TAGGING_BERT
from .._models._xlnet_model import TAGGING_XLNET


def transformer(
    path, s3_path, class_name, model = 'xlnet', size = 'base', validate = True
):
    if validate:
        check_file(path[model][size], s3_path[model][size])
    else:
        if not check_available(path[model][size]):
            raise Exception(
                '%s/%s/%s is not available, please `validate = True`'
                % (class_name, model, size)
            )

    try:
        with open(path[model][size]['setting']) as fopen:
            nodes = json.load(fopen)
        g = load_graph(path[model][size]['model'])
    except:
        raise Exception(
            "model corrupted due to some reasons, please run malaya.clear_cache('%s/%s/%s') and try again"
            % (class_name, model, size)
        )

    if model in ['albert', 'bert']:
        tokenizer, cls, sep = sentencepiece_tokenizer_bert(
            path[model][size]['tokenizer'], path[model][size]['vocab']
        )
        return TAGGING_BERT(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = None,
            input_masks = None,
            logits = g.get_tensor_by_name('import/logits:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            cls = cls,
            sep = sep,
            settings = nodes,
        )

    if model in ['xlnet']:
        tokenizer = sentencepiece_tokenizer_xlnet(
            path[model][size]['tokenizer']
        )
        return TAGGING_XLNET(
            X = g.get_tensor_by_name('import/Placeholder:0'),
            segment_ids = g.get_tensor_by_name('import/Placeholder_1:0'),
            input_masks = g.get_tensor_by_name('import/Placeholder_2:0'),
            logits = g.get_tensor_by_name('import/logits:0'),
            sess = generate_session(graph = g),
            tokenizer = tokenizer,
            settings = nodes,
        )



import subprocess

files = '../parliament-text.txt,../wiki-text.txt,../dumping-twitter.txt,../news-text.txt,../dumping-instagram.txt'.split(
    ','
)

for no, file in enumerate(files):
    print('Reading from input files', file)

    output_files = 'albert-%d.tfrecord' % (no)
    print('Output filename', output_files)
    subprocess.call(
        'python3 create_pretraining_data.py --input_file=%s --output_file=%s --vocab_file=sp10m.cased.v8.vocab --spm_model_file=sp10m.cased.v8.model --do_lower_case=False --dupe_factor=5'
        % (file, output_files),
        shell = True,
    )



import numpy as np
import collections
from sklearn.utils import shuffle
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import TruncatedSVD, NMF, LatentDirichletAllocation
from .stem import sastrawi
from ._models._lda2vec import LDA2VEC
from .texts._text_functions import (
    simple_textcleaning,
    STOPWORDS,
    classification_textcleaning,
    print_topics_modelling,
    build_dataset,
)
from .texts.vectorizer import skipgrams, SkipGramVectorizer
from .generator import ngrams as ngrams_generator


def _softmax(x):
    e_x = np.exp(x - np.max(x))
    out = e_x / e_x.sum()
    return out


def _softmax_2d(x):
    y = x - x.max(axis = 1, keepdims = True)
    np.exp(y, out = y)
    y /= y.sum(axis = 1, keepdims = True)
    return y


def _prob_words(context, vocab, temperature = 1.0):
    dot = np.dot(vocab, context)
    prob = _softmax(dot / temperature)
    return prob


def _prepare_topics(
    weights,
    factors,
    word_vectors,
    vocab,
    temperature = 1.0,
    doc_lengths = None,
    term_frequency = None,
    normalize = False,
):
    topic_to_word = []
    msg = 'Vocabulary size did not match size of word vectors'
    if not len(vocab) == word_vectors.shape[0]:
        raise ValueError(msg)
    if normalize:
        word_vectors /= np.linalg.norm(word_vectors, axis = 1)[:, None]
    for factor_vector in factors:
        factor_to_word = _prob_words(
            factor_vector, word_vectors, temperature = temperature
        )
        topic_to_word.append(np.ravel(factor_to_word))
    topic_to_word = np.array(topic_to_word)
    msg = 'Not all rows in topic_to_word sum to 1'
    if not np.allclose(np.sum(topic_to_word, axis = 1), 1):
        raise ValueError(msg)
    doc_to_topic = _softmax_2d(weights)
    msg = 'Not all rows in doc_to_topic sum to 1'
    if not np.allclose(np.sum(doc_to_topic, axis = 1), 1):
        raise ValueError(msg)
    data = {
        'topic_term_dists': topic_to_word,
        'doc_topic_dists': doc_to_topic,
        'doc_lengths': doc_lengths,
        'vocab': vocab,
        'term_frequency': term_frequency,
    }
    return data


class _ATTENTION_TOPIC:
    def __init__(self, features, components):
        self._features = features
        self._components = components

    def top_topics(self, len_topic, top_n = 10, return_df = True):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        len_topic: int
        """
        if not isinstance(len_topic, int):
            raise ValueError('len_topic must be an integer')
        if not isinstance(top_n, int):
            raise ValueError('top_n must be an integer')
        if not isinstance(return_df, bool):
            raise ValueError('return_df must be a boolean')
        return print_topics_modelling(
            len_topic,
            feature_names = np.array(self._features),
            sorting = np.argsort(self._components)[:, ::-1],
            n_words = top_n,
            return_df = return_df,
        )

    def get_topics(self, len_topic):
        """
        Return important topics based on decomposition.

        Parameters
        ----------
        len_topic: int

        Returns
        -------
        results: list of strings
        """
        if not isinstance(len_topic, int):
            raise ValueError('len_topic must be an integer')
        results = []
        for no, topic in enumerate(self._components):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self._features[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results


class _DEEP_TOPIC:
    def __init__(
        self,
        model,
        dictionary,
        reversed_dictionary,
        freqs,
        len_idx_text_clean,
        corpus,
    ):
        self._model = model
        self._dictionary = dictionary
        self._reversed_dictionary = reversed_dictionary
        self._doc_embed = model.sess.run(model.doc_embedding)
        self._topic_embed = model.sess.run(model.topic_embedding)
        self._word_embed = model.sess.run(model.word_embedding)
        self._components = self._topic_embed.dot(self._word_embed.T)
        self._freqs = freqs
        self._doc_len = len_idx_text_clean
        self._corpus = corpus
        self._features = []
        for i in range(len(self._dictionary)):
            self._features.append(self._reversed_dictionary[i])

    def visualize_topics(self, notebook_mode = False, mds = 'pcoa'):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        mds : str, optional (default='pcoa')
            2D Decomposition. Allowed values:

            * ``'pcoa'`` - Dimension reduction via Jensen-Shannon Divergence & Principal Coordinate Analysis (aka Classical Multidimensional Scaling)
            * ``'mmds'`` - Dimension reduction via Multidimensional scaling
            * ``'tsne'`` - Dimension reduction via t-distributed stochastic neighbor embedding
        """
        if not isinstance(mds, str):
            raise ValueError('mds must be a string')
        if not isinstance(notebook_mode, bool):
            raise ValueError('notebook_mode must be a boolean')
        try:
            import pyLDAvis
            import pyLDAvis.sklearn
        except:
            raise Exception(
                'pyldavis not installed. Please install it and try again.'
            )

        if notebook_mode:
            pyLDAvis.enable_notebook()

        vis_data = _prepare_topics(
            self._doc_embed,
            self._topic_embed,
            self._word_embed,
            np.array(self._features),
            doc_lengths = self._doc_len,
            term_frequency = self._freqs,
            normalize = True,
        )
        prepared_vis_data = pyLDAvis.prepare(**vis_data)
        if notebook_mode:
            return prepared_vis_data
        else:
            pyLDAvis.show(prepared_vis_data)

    def top_topics(self, len_topic, top_n = 10, return_df = True):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        len_topic: int
        """
        if not isinstance(len_topic, int):
            raise ValueError('len_topic must be an integer')
        if not isinstance(top_n, int):
            raise ValueError('top_n must be an integer')
        if not isinstance(return_df, bool):
            raise ValueError('return_df must be a boolean')
        return print_topics_modelling(
            len_topic,
            feature_names = np.array(self._features),
            sorting = np.argsort(self._components)[:, ::-1],
            n_words = top_n,
            return_df = return_df,
        )

    def get_topics(self, len_topic):
        """
        Return important topics based on decomposition.

        Parameters
        ----------
        len_topic: int

        Returns
        -------
        results: list of strings
        """
        if not isinstance(len_topic, int):
            raise ValueError('len_topic must be an integer')
        results = []
        for no, topic in enumerate(self._components):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self._features[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results

    def get_sentences(self, len_sentence, k = 0):
        """
        Return important sentences related to selected column based on decomposition.

        Parameters
        ----------
        len_sentence: int
        k: int, (default=0)
            index of decomposition matrix.

        Returns
        -------
        results: list of strings
        """
        if not isinstance(len_sentence, int):
            raise ValueError('len_sentence must be an integer')
        if not isinstance(k, int):
            raise ValueError('k must be an integer')
        if not (k < self._doc_embed.shape[1] and k >= 0):
            raise ValueError('k should be between 0 and n_topics')
        reverse_sorted = np.argsort(self._doc_embed[:, k])[::-1]
        return [self._corpus[i] for i in reverse_sorted[:len_sentence]]


class _TOPIC:
    def __init__(
        self, features, comp, corpus, transformed, vectorizer, vectors
    ):
        self.features = features
        self.comp = comp
        self.corpus = corpus
        self.transformed = transformed
        self.vectorizer = vectorizer
        self._vectors = vectors

    def visualize_topics(self, notebook_mode = False, mds = 'pcoa'):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        mds : str, optional (default='pcoa')
            2D Decomposition. Allowed values:

            * ``'pcoa'`` - Dimension reduction via Jensen-Shannon Divergence & Principal Coordinate Analysis (aka Classical Multidimensional Scaling)
            * ``'mmds'`` - Dimension reduction via Multidimensional scaling
            * ``'tsne'`` - Dimension reduction via t-distributed stochastic neighbor embedding
        """
        if not isinstance(mds, str):
            raise ValueError('mds must be a string')
        if not isinstance(notebook_mode, bool):
            raise ValueError('notebook_mode must be a boolean')
        if not isinstance(self.comp, LatentDirichletAllocation):
            raise ValueError('only support lda_topic_modelling()')

        import pyLDAvis
        import pyLDAvis.sklearn

        if notebook_mode:
            pyLDAvis.enable_notebook()

        prepared_vis_data = pyLDAvis.sklearn.prepare(
            self.comp, self._vectors, self.vectorizer, mds = mds
        )
        if notebook_mode:
            return prepared_vis_data
        else:
            pyLDAvis.show(prepared_vis_data)

    def top_topics(self, len_topic, top_n = 10, return_df = True):
        """
        Print important topics based on decomposition.

        Parameters
        ----------
        len_topic: int
        """
        if not isinstance(len_topic, int):
            raise ValueError('len_topic must be an integer')
        if not isinstance(return_df, bool):
            raise ValueError('return_df must be a boolean')
        return print_topics_modelling(
            len_topic,
            feature_names = np.array(self.features),
            sorting = np.argsort(self.comp.components_)[:, ::-1],
            n_words = top_n,
            return_df = return_df,
        )

    def get_topics(self, len_topic):
        """
        Return important topics based on decomposition.

        Parameters
        ----------
        len_topic: int

        Returns
        -------
        results: list of strings
        """
        if not isinstance(len_topic, int):
            raise ValueError('len_topic must be an integer')
        results = []
        for no, topic in enumerate(self.comp.components_):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self.features[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results

    def get_sentences(self, len_sentence, k = 0):
        """
        Return important sentences related to selected column based on decomposition.

        Parameters
        ----------
        len_sentence: int
        k: int, (default=0)
            index of decomposition matrix.

        Returns
        -------
        results: list of strings
        """
        if not isinstance(len_sentence, int):
            raise ValueError('len_sentence must be an integer')
        if not isinstance(k, int):
            raise ValueError('k must be an integer')
        if not (k < self.transformed.shape[1] and k >= 0):
            raise ValueError('k should be between 0 and n_topics')
        reverse_sorted = np.argsort(self.transformed[:, k])[::-1]
        return [self.corpus[i] for i in reverse_sorted[:len_sentence]]


def _base_topic_modelling(
    corpus,
    n_topics,
    decomposition,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    vectorizer = 'bow',
    stemming = sastrawi,
    cleaning = simple_textcleaning,
    stop_words = None,
    **kwargs,
):
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(n_topics, int):
        raise ValueError('n_topics must be an integer')
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(stemming, collections.Callable) and stemming is not None:
        raise ValueError('stemming must be a callable type or None')
    if not isinstance(cleaning, collections.Callable) and stemming is not None:
        raise ValueError('cleaning must be a callable type or None')
    if (
        not isinstance(stop_words, set)
        and not isinstance(stop_words, list)
        and stop_words is not None
    ):
        raise ValueError('stop_words must be a set or a list or None')
    vectorizer = vectorizer.lower()
    if not vectorizer in ['tfidf', 'bow', 'skip-gram']:
        raise ValueError("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not (isinstance(max_df, int) or isinstance(max_df, float)):
        raise ValueError('max_df must be an integer or a float')
    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df <= 1 and max_df > 0):
        raise ValueError(
            'max_df must be bigger than 0, less than or equal to 1'
        )
    if len(corpus) < n_topics:
        raise ValueError(
            'length corpus must be bigger than or equal to n_topics'
        )

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming is not None:
        for i in range(len(corpus)):
            corpus[i] = stemming(corpus[i])
    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    tf_vectorizer = Vectorizer(
        max_df = max_df,
        min_df = min_df,
        ngram_range = ngram,
        stop_words = stop_words,
        **kwargs,
    )
    tf = tf_vectorizer.fit_transform(corpus)
    tf_features = tf_vectorizer.get_feature_names()
    compose = decomposition(n_topics).fit(tf)
    return _TOPIC(
        tf_features,
        compose,
        [classification_textcleaning(c) for c in corpus],
        compose.transform(tf),
        tf_vectorizer,
        tf,
    )


def lda(
    corpus,
    n_topics = 10,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    stemming = sastrawi,
    vectorizer = 'bow',
    cleaning = simple_textcleaning,
    stop_words = None,
    **kwargs,
):
    """
    Train a LDA model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    stemming: function, (default=sastrawi)
        function to stem the corpus.
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word.
        * ``'tfidf'`` - Term frequency inverse Document Frequency.
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS

    Returns
    -------
    _TOPIC: malaya.topic_modelling._TOPIC class
    """
    if stop_words is None:
        stop_words = STOPWORDS
    return _base_topic_modelling(
        corpus,
        n_topics,
        LatentDirichletAllocation,
        max_df = max_df,
        min_df = min_df,
        ngram = ngram,
        vectorizer = vectorizer,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
        **kwargs,
    )


def nmf(
    corpus,
    n_topics = 10,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    stemming = sastrawi,
    vectorizer = 'bow',
    cleaning = simple_textcleaning,
    stop_words = None,
    **kwargs,
):
    """
    Train a NMF model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    stemming: function, (default=sastrawi)
        function to stem the corpus.
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word.
        * ``'tfidf'`` - Term frequency inverse Document Frequency.
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS

    Returns
    -------
    _TOPIC: malaya.topic_modelling._TOPIC class
    """
    if stop_words is None:
        stop_words = STOPWORDS
    return _base_topic_modelling(
        corpus,
        n_topics,
        NMF,
        max_df = max_df,
        min_df = min_df,
        ngram = ngram,
        vectorizer = vectorizer,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
        **kwargs,
    )


def lsa(
    corpus,
    n_topics,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    vectorizer = 'bow',
    stemming = sastrawi,
    cleaning = simple_textcleaning,
    stop_words = None,
    **kwargs,
):
    """
    Train a LSA model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word.
        * ``'tfidf'`` - Term frequency inverse Document Frequency.
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams.
    stemming: function, (default=sastrawi)
        function to stem the corpus.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS

    Returns
    -------
    _TOPIC: malaya.topic_modelling._TOPIC class
    """
    if stop_words is None:
        stop_words = STOPWORDS
    return _base_topic_modelling(
        corpus,
        n_topics,
        TruncatedSVD,
        max_df = max_df,
        min_df = min_df,
        ngram = ngram,
        vectorizer = vectorizer,
        stemming = stemming,
        cleaning = cleaning,
        stop_words = stop_words,
        **kwargs,
    )


def lda2vec(
    corpus,
    n_topics,
    stemming = sastrawi,
    max_df = 0.95,
    min_df = 2,
    ngram = (1, 3),
    cleaning = simple_textcleaning,
    vectorizer = 'bow',
    stop_words = None,
    window_size = 2,
    embedding_size = 128,
    epoch = 10,
    switch_loss = 3,
    skip = 5,
    **kwargs,
):
    """
    Train a LDA2Vec model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    stemming: function, (default=sastrawi)
        function to stem the corpus.
    max_df: float, (default=0.95)
        maximum of a word selected based on document frequency.
    min_df: int, (default=2)
        minimum of a word selected on based on document frequency.
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS
    embedding_size: int, (default=128)
        embedding size of lda2vec tensors.
    training_iteration: int, (default=10)
        training iteration, how many loop need to train.
    switch_loss: int, (default=3)
        baseline to switch from document based loss to document + word based loss.
    vectorizer: str, (default='bow')
        vectorizer technique. Allowed values:

        * ``'bow'`` - Bag of Word.
        * ``'tfidf'`` - Term frequency inverse Document Frequency.
        * ``'skip-gram'`` - Bag of Word with skipping certain n-grams.
    skip: int, (default=5)
        skip value if vectorizer = 'skip-gram'

    Returns
    -------
    _DEEP_TOPIC: malaya.topic_modelling._DEEP_TOPIC class
    """
    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(n_topics, int):
        raise ValueError('n_topics must be an integer')
    if not isinstance(vectorizer, str):
        raise ValueError('vectorizer must be a string')
    if not isinstance(stemming, collections.Callable) and stemming is not None:
        raise ValueError('stemming must be a callable type or None')
    if not isinstance(cleaning, collections.Callable) and stemming is not None:
        raise ValueError('cleaning must be a callable type or None')
    if (
        not isinstance(stop_words, set)
        and not isinstance(stop_words, list)
        and stop_words is not None
    ):
        raise ValueError('stop_words must be a set or a list or None')
    vectorizer = vectorizer.lower()
    if not vectorizer in ['tfidf', 'bow', 'skip-gram']:
        raise ValueError("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not isinstance(min_df, int):
        raise ValueError('min_df must be an integer')
    if not (isinstance(max_df, int) or isinstance(max_df, float)):
        raise ValueError('max_df must be an integer or a float')
    if min_df < 1:
        raise ValueError('min_df must be bigger than 0')
    if not (max_df <= 1 and max_df > 0):
        raise ValueError(
            'max_df must be bigger than 0, less than or equal to 1'
        )
    if not isinstance(embedding_size, int):
        raise ValueError('embedding_size must be an integer')
    if not isinstance(window_size, int):
        raise ValueError('window_size must be an integer')
    if not isinstance(epoch, int):
        raise ValueError('epoch must be an integer')
    if not isinstance(switch_loss, int):
        raise ValueError('switch_loss must be an integer')
    if not isinstance(skip, int):
        raise ValueError('skip must be an integer')

    if vectorizer == 'tfidf':
        Vectorizer = TfidfVectorizer
    elif vectorizer == 'bow':
        Vectorizer = CountVectorizer
    elif vectorizer == 'skip-gram':
        Vectorizer = SkipGramVectorizer
    else:
        raise Exception("vectorizer must be in  ['tfidf', 'bow', 'skip-gram']")
    tf_vectorizer = Vectorizer(
        ngram_range = ngram,
        min_df = min_df,
        max_df = max_df,
        stop_words = stop_words,
    )
    if stop_words is None:
        stop_words = STOPWORDS

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming is not None:
        for i in range(len(corpus)):
            corpus[i] = stemming(corpus[i])
    text_clean = []
    for text in corpus:
        text_clean.append(
            ' '.join([word for word in text.split() if word not in stop_words])
        )
    tf_vectorizer.fit(text_clean)
    idx_text_clean, len_idx_text_clean = [], []
    transformed_text_clean = tf_vectorizer.transform(text_clean)
    for text in transformed_text_clean:
        splitted = text.nonzero()[1]
        idx_text_clean.append(splitted)
        len_idx_text_clean.append(len(splitted))
    dictionary = {
        i: no for no, i in enumerate(tf_vectorizer.get_feature_names())
    }
    reversed_dictionary = {
        no: i for no, i in enumerate(tf_vectorizer.get_feature_names())
    }
    freqs = transformed_text_clean.toarray().sum(axis = 0).tolist()
    doc_ids = np.arange(len(idx_text_clean))
    num_unique_documents = doc_ids.max()
    pivot_words, target_words, doc_ids = [], [], []
    for i, t in enumerate(idx_text_clean):
        pairs, _ = skipgrams(
            t,
            vocabulary_size = len(dictionary),
            window_size = window_size,
            shuffle = True,
            negative_samples = 0,
        )
        for pair in pairs:
            temp_data = pair
            pivot_words.append(temp_data[0])
            target_words.append(temp_data[1])
            doc_ids.append(i)
    pivot_words, target_words, doc_ids = shuffle(
        pivot_words, target_words, doc_ids, random_state = 10
    )
    num_unique_documents = len(idx_text_clean)

    model = LDA2VEC(
        num_unique_documents,
        len(dictionary),
        n_topics,
        freqs,
        embedding_size = embedding_size,
        **kwargs,
    )
    model.train(
        pivot_words, target_words, doc_ids, epoch, switch_loss = switch_loss
    )
    return _DEEP_TOPIC(
        model,
        dictionary,
        reversed_dictionary,
        freqs,
        len_idx_text_clean,
        text_clean,
    )


def attention(
    corpus,
    n_topics,
    vectorizer,
    stemming = sastrawi,
    cleaning = simple_textcleaning,
    stop_words = None,
    ngram = (1, 3),
    batch_size = 10,
):

    """
    Use attention from vectorizer model to do topic modelling based on corpus / list of strings given.

    Parameters
    ----------
    corpus: list
    n_topics: int, (default=10)
        size of decomposition column.
    vectorizer: object
    stemming: function, (default=sastrawi)
        function to stem the corpus.
    cleaning: function, (default=simple_textcleaning)
        function to clean the corpus.
    stop_words: list, (default=None)
        list of stop words to remove. If None, default is malaya.texts._text_functions.STOPWORDS
    ngram: tuple, (default=(1,3))
        n-grams size to train a corpus.
    batch_size: int, (default=10)
        size of strings for each vectorization and attention.

    Returns
    -------
    _ATTENTION_TOPIC: malaya.topic_modelling._ATTENTION_TOPIC class
    """

    if not isinstance(corpus, list):
        raise ValueError('corpus must be a list')
    if not isinstance(corpus[0], str):
        raise ValueError('corpus must be list of strings')
    if not isinstance(n_topics, int):
        raise ValueError('n_topics must be an integer')
    if not isinstance(batch_size, int):
        raise ValueError('batch_size must be an integer')
    if not isinstance(ngram, tuple):
        raise ValueError('ngram must be a tuple')
    if not len(ngram) == 2:
        raise ValueError('ngram size must equal to 2')
    if not hasattr(vectorizer, 'attention') and not hasattr(
        vectorizer, 'vectorize'
    ):
        raise ValueError(
            'vectorizer must has `attention` and `vectorize` methods'
        )
    if not isinstance(stemming, collections.Callable) and stemming is not None:
        raise ValueError('stemming must be a callable type or None')
    if not isinstance(cleaning, collections.Callable) and stemming is not None:
        raise ValueError('cleaning must be a callable type or None')
    if (
        not isinstance(stop_words, set)
        and not isinstance(stop_words, list)
        and stop_words is not None
    ):
        raise ValueError('stop_words must be a set or a list or None')
    if len(corpus) < n_topics:
        raise ValueError(
            'length corpus must be bigger than or equal to n_topics'
        )

    from sklearn.cluster import KMeans

    if stop_words is None:
        stop_words = STOPWORDS

    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming is not None:
        for i in range(len(corpus)):
            corpus[i] = stemming(corpus[i])

    def generate_ngram(seq, ngram = (1, 3)):
        g = []
        for i in range(ngram[0], ngram[-1] + 1):
            g.extend(list(ngrams_generator(seq, i)))
        return g

    rows, attentions = [], []
    for i in range(0, len(corpus), batch_size):
        index = min(i + batch_size, len(corpus))
        rows.append(vectorizer.vectorize(corpus[i:index]))
        attentions.extend(vectorizer.attention(corpus[i:index]))

    concat = np.concatenate(rows, axis = 0)
    kmeans = KMeans(n_clusters = n_topics, random_state = 0).fit(concat)
    labels = kmeans.labels_

    overall, filtered_a = [], []
    for a in attentions:
        f = [i for i in a if i[0] not in stop_words]
        overall.extend(f)
        filtered_a.append(f)

    o_ngram = generate_ngram(overall, ngram)
    features = []
    for i in o_ngram:
        features.append(' '.join([w[0] for w in i]))
    features = list(set(features))

    components = np.zeros((n_topics, len(features)))
    for no, i in enumerate(labels):
        f = generate_ngram(filtered_a[no], ngram)
        for w in f:
            word = ' '.join([r[0] for r in w])
            score = np.mean([r[1] for r in w])
            if word in features:
                components[i, features.index(word)] += score

    return _ATTENTION_TOPIC(features, components)




# coding: utf-8

# In[1]:


import word2vec
import numpy as np
import tensorflow as tf
import json
import os
import re
from unidecode import unidecode
os.environ['CUDA_VISIBLE_DEVICES'] = ''


# In[2]:


with open('wiki-ms.txt') as fopen:
    sentences = fopen.read()


def cleaning(string):
    string = re.sub(
        'http\S+|www.\S+',
        '',
        ' '.join(
            [i for i in string.split() if i.find('#') < 0 and i.find('@') < 0]
        ),
    )
    string = unidecode(string).replace('.', '. ').replace(',', ', ')
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = re.sub(r'[ ]+', ' ', string).strip()
    string = ' '.join(
            [
                i
                for i in re.findall('[\\w\']+|[;:\-\(\)&.,!?"]', string)
                if len(i)
            ]
        )
    return string.lower()
# In[3]:

sentences = cleaning(sentences).split()

word_array, dictionary, rev_dictionary, num_lines, num_words = word2vec.build_word_array(sentences,vocab_size=1000000)


# In[4]:


len(dictionary)


# In[5]:


X, Y = word2vec.build_training_set(word_array)
graph_params = {'batch_size': 32,
                'vocab_size': np.max(X)+1,
                'embed_size': 256,
                'hid_size': 256,
                'neg_samples': 128,
                'learn_rate': 0.01,
                'momentum': 0.9,
                'embed_noise': 0.1,
                'hid_noise': 0.3,
                'epoch':10,
                'optimizer': 'Momentum'}


# In[6]:


split = round(X.shape[0]*0.9)
train_X, train_Y = X[:split, :], Y[:split, :]
test_X, test_Y = X[split:, :], Y[split:, :]


# In[7]:


model = word2vec.Model(graph_params)
print('model built, vocab size %d, document length %d'%(np.max(X)+1, len(word_array)))


# In[ ]:


embed_weights, nce_weights = model.train(train_X, train_Y, test_X, test_Y,
                                         graph_params['epoch'],
                                         graph_params['batch_size'])


# In[ ]:


import pickle
with open('word2vec-wiki-256.p', 'wb') as fopen:
    pickle.dump({'dictionary':dictionary,'rev_dictionary':rev_dictionary,
                 'embed_weights':embed_weights,'nce_weights':nce_weights}, fopen)




import tensorflow as tf
import numpy as np
from ..texts._text_functions import (
    str_idx,
    entities_textcleaning,
    char_str_idx,
    generate_char_seq,
    language_detection_textcleaning,
    tag_chunk,
)
from .._utils._parse_dependency import DependencyGraph
from .._utils._utils import add_neutral as neutral
from .._utils._html import (
    _render_binary,
    _render_toxic,
    _render_emotion,
    _render_relevancy,
)


def _convert_sparse_matrix_to_sparse_tensor(X, got_limit = True, limit = 5):
    coo = X.tocoo()
    indices = np.mat([coo.row, coo.col]).transpose()
    if got_limit:
        coo.data[coo.data > limit] = limit
    return (
        tf.SparseTensorValue(indices, coo.col, coo.shape),
        tf.SparseTensorValue(indices, coo.data, coo.shape),
    )


class _LANG_MODEL:
    def __init__(self):
        self.X = tf.sparse_placeholder(tf.int32)
        self.W = tf.sparse_placeholder(tf.int32)
        self.Y = tf.placeholder(tf.int32, [None])
        embeddings = tf.Variable(tf.truncated_normal([660726, 40]))
        embed = tf.nn.embedding_lookup_sparse(
            embeddings, self.X, self.W, combiner = 'mean'
        )
        self.logits = tf.layers.dense(embed, 4)


class _SPARSE_SOFTMAX_MODEL:
    def __init__(self, output_size, embedded_size, vocab_size):
        self.X = tf.sparse_placeholder(tf.int32)
        self.W = tf.sparse_placeholder(tf.int32)
        self.Y = tf.placeholder(tf.int32, [None])
        embeddings = tf.Variable(
            tf.truncated_normal([vocab_size, embedded_size])
        )
        embed = tf.nn.embedding_lookup_sparse(
            embeddings, self.X, self.W, combiner = 'mean'
        )
        self.logits = tf.layers.dense(embed, output_size)


class DEEP_LANG:
    def __init__(self, path, vectorizer, label):
        self._graph = tf.Graph()
        with self._graph.as_default():
            self._model = _LANG_MODEL()
            self._sess = tf.InteractiveSession()
            self._sess.run(tf.global_variables_initializer())
            saver = tf.train.Saver(tf.trainable_variables())
            saver.restore(self._sess, path + '/model.ckpt')
        self._vectorizer = vectorizer
        self._label = label
        self._softmax = tf.nn.softmax(self._model.logits)

    def predict(self, string, get_proba = False):
        """
        classify a string.

        Parameters
        ----------
        string : str

        Returns
        -------
        dictionary: results
        """
        if not isinstance(string, str):
            raise ValueError('input must be a string')
        if not isinstance(get_proba, bool):
            raise ValueError('get_proba must be a boolean')

        string = language_detection_textcleaning(string)
        transformed = self._vectorizer.transform([string])
        batch_x = _convert_sparse_matrix_to_sparse_tensor(transformed)
        probs = self._sess.run(
            self._softmax,
            feed_dict = {self._model.X: batch_x[0], self._model.W: batch_x[1]},
        )[0]
        if get_proba:
            return {self._label[no]: i for no, i in enumerate(probs)}
        else:
            return self._label[np.argmax(probs)]

    def predict_batch(self, strings, get_proba = False):
        """
        classify list of strings

        Parameters
        ----------
        strings : list

        Returns
        -------
        list_dictionaries: list of results
        """
        if not isinstance(strings, list):
            raise ValueError('input must be a list')
        if not isinstance(strings[0], str):
            raise ValueError('input must be list of strings')
        if not isinstance(get_proba, bool):
            raise ValueError('get_proba must be a boolean')

        strings = [language_detection_textcleaning(i) for i in strings]
        transformed = self._vectorizer.transform(strings)
        batch_x = _convert_sparse_matrix_to_sparse_tensor(transformed)
        probs = self._sess.run(
            self._softmax,
            feed_dict = {self._model.X: batch_x[0], self._model.W: batch_x[1]},
        )
        dicts = []
        if get_proba:
            for i in range(probs.shape[0]):
                dicts.append(
                    {self._label[no]: k for no, k in enumerate(probs[i])}
                )
        else:
            probs = np.argmax(probs, 1)
            for prob in probs:
                dicts.append(self._label[prob])
        return dicts



# Copyright (c) 2018 luozhouyang, https://github.com/luozhouyang/python-string-similarity/blob/master/strsimpy/jaro_winkler.py
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class StringDistance:
    def distance(self, s0, s1):
        raise NotImplementedError()


class NormalizedStringDistance(StringDistance):
    def distance(self, s0, s1):
        raise NotImplementedError()


class MetricStringDistance(StringDistance):
    def distance(self, s0, s1):
        raise NotImplementedError()


class StringSimilarity:
    def similarity(self, s0, s1):
        raise NotImplementedError()


class NormalizedStringSimilarity(StringSimilarity):
    def similarity(self, s0, s1):
        raise NotImplementedError()


class JaroWinkler(NormalizedStringSimilarity, NormalizedStringDistance):
    def __init__(self, threshold = 0.7):
        self.threshold = threshold
        self.three = 3
        self.jw_coef = 0.1

    def get_threshold(self):
        return self.threshold

    def similarity(self, s0, s1):
        if s0 is None:
            raise TypeError('Argument s0 is NoneType.')
        if s1 is None:
            raise TypeError('Argument s1 is NoneType.')
        if s0 == s1:
            return 1.0
        mtp = self.matches(s0, s1)
        m = mtp[0]
        if m == 0:
            return 0.0
        j = (m / len(s0) + m / len(s1) + (m - mtp[1]) / m) / self.three
        jw = j
        if j > self.get_threshold():
            jw = j + min(self.jw_coef, 1.0 / mtp[self.three]) * mtp[2] * (1 - j)
        return jw

    def distance(self, s0, s1):
        return 1.0 - self.similarity(s0, s1)

    @staticmethod
    def matches(s0, s1):
        if len(s0) > len(s1):
            max_str = s0
            min_str = s1
        else:
            max_str = s1
            min_str = s0
        ran = int(max(len(max_str) / 2 - 1, 0))
        match_indexes = [-1] * len(min_str)
        match_flags = [False] * len(max_str)
        matches = 0
        for mi in range(len(min_str)):
            c1 = min_str[mi]
            for xi in range(max(mi - ran, 0), min(mi + ran + 1, len(max_str))):
                if not match_flags[xi] and c1 == max_str[xi]:
                    match_indexes[mi] = xi
                    match_flags[xi] = True
                    matches += 1
                    break

        ms0, ms1 = [0] * matches, [0] * matches
        si = 0
        for i in range(len(min_str)):
            if match_indexes[i] != -1:
                ms0[si] = min_str[i]
                si += 1
        si = 0
        for j in range(len(max_str)):
            if match_flags[j]:
                ms1[si] = max_str[j]
                si += 1
        transpositions = 0
        for mi in range(len(ms0)):
            if ms0[mi] != ms1[mi]:
                transpositions += 1
        prefix = 0
        for mi in range(len(min_str)):
            if s0[mi] == s1[mi]:
                prefix += 1
            else:
                break
        return [matches, int(transpositions / 2), prefix, len(max_str)]



from collections import Counter, defaultdict
from itertools import product
import numpy as np
import json
import re
from .texts._jarowrinkler import JaroWinkler
from .texts._text_functions import ENGLISH_WORDS, MALAY_WORDS
from .texts._tatabahasa import (
    alphabet,
    consonants,
    vowels,
    rules_normalizer,
    permulaan,
    hujung,
    stopword_tatabahasa,
    quad_vowels,
    group_compound,
)
from ._utils._paths import PATH_NGRAM, S3_PATH_NGRAM
from ._utils._utils import check_file, check_available


def _load_sentencepiece(vocab, vocab_model):
    import sentencepiece as spm
    from .texts._text_functions import SentencePieceTokenizer

    sp_model = spm.SentencePieceProcessor()
    sp_model.Load(vocab_model)

    with open(vocab) as fopen:
        v = fopen.read().split('\n')[:-1]
    v = [i.split('\t') for i in v]
    v = {i[0]: i[1] for i in v}
    tokenizer = SentencePieceTokenizer(v, sp_model)
    return tokenizer


def tokens_to_masked_ids(tokens, mask_ind, tokenizer):
    masked_tokens = tokens[:]
    masked_tokens[mask_ind] = '<mask>'
    masked_tokens = ['<cls>'] + masked_tokens + ['<sep>']
    masked_ids = tokenizer.convert_tokens_to_ids(masked_tokens)
    return masked_ids


def generate_ids(mask, tokenizer):
    tokens = tokenizer.tokenize(mask)
    input_ids = [
        tokens_to_masked_ids(tokens, i, tokenizer) for i in range(len(tokens))
    ]
    tokens_ids = tokenizer.convert_tokens_to_ids(tokens)
    return tokens, input_ids, tokens_ids


def _build_dicts(words):
    occurences = {}
    for l in alphabet:
        occurences[l] = defaultdict(lambda: 0)
    for word in words:
        occurences[word[0]][word] += 1
    return occurences


def _get_indices(string, c = 'a'):
    return [i for i in range(len(string)) if string[i] == c]


def _permutate(string, indices):
    p = [''.join(_set) for _set in product(list(vowels), repeat = len(indices))]
    p = [p_ for p_ in p if not all([a in p_ for a in quad_vowels])]
    mutate = []
    for p_ in p:
        s = list(string[:])
        for i in range(len(indices)):
            s[indices[i]] = p_[i]
        mutate.append(''.join(s))
    return mutate


def _permutate_sp(string, indices, sp_tokenizer):
    p = [''.join(_set) for _set in product(list(vowels), repeat = len(indices))]
    p = [p_ for p_ in p if not all([a in p_ for a in quad_vowels])]
    mutate = []
    for p_ in p:
        s = list(string[:])
        for i in range(len(indices)):
            s[indices[i]] = p_[i]
        s = ''.join(s)
        if sp_tokenizer.tokenize(s)[0] == '▁':
            continue
        mutate.append(s)
    return mutate


def _augment_vowel_alternate(string):
    r = []
    # a flag to not duplicate
    last_time = False
    for i, c in enumerate(string[:-1], 1):
        last = i - 2
        if last < 0:
            last = 0

        # we only want to put a vowel after consonant if next that consonant if not a wovel
        if c in consonants and string[i] not in vowels:
            if c + string[i] in group_compound and not last_time:
                r.append(c + string[i])
                last_time = True
            elif string[last] + c in group_compound and not last_time:
                r.append(string[last] + c)
                last_time = True
            else:
                last_time = False
                if len(r):
                    # ['ng'] gg
                    if (
                        r[-1] in group_compound
                        and c + string[i] == r[-1][-1] * 2
                    ):
                        r.append('^')
                        continue
                    elif r[-1] in group_compound and c == r[-1][-1]:
                        if c + string[i] in group_compound:
                            continue
                        else:
                            r.append('a')
                            continue
                r.append(c + 'a')

        else:
            if len(r):
                if r[-1] in group_compound and c == r[-1][-1]:
                    continue
            r.append(c)

    if r[-1][-1] in vowels and string[-1] in consonants:
        r.append(string[-1])

    elif (
        r[-1] in group_compound
        and string[-2] in vowels
        and string[-1] in consonants
    ):
        r.append(string[-2:])

    left = ''.join(r).replace('^', '')
    right = left + 'a'
    return left, right


def _augment_vowel_prob(word, **kwargs):
    l, r = _augment_vowel_alternate(word)
    return list(
        set(_permutate(l, _get_indices(l)) + _permutate(r, _get_indices(r)))
    )


def _augment_vowel_prob_sp(word, sp_tokenizer):
    l, r = _augment_vowel_alternate(word)
    return list(
        set(
            _permutate_sp(l, _get_indices(l), sp_tokenizer)
            + _permutate_sp(r, _get_indices(r), sp_tokenizer)
        )
    )


def _augment_vowel(
    string, selected = ['a', 'u', 'i', 'e'], included_end = True
):
    pseudo = []
    if included_end:
        end = string[-1]
    else:
        end = ''
    for c in selected:
        pseudo.append(''.join([w + c for w in string[:-1]] + [end]))
    return pseudo


def _return_possible(word, dicts, edits):
    return set(e2 for e1 in edits(word) for e2 in edits(e1) if e2 in dicts)


def _return_known(word, dicts):
    return set(w for w in word if w in dicts)


class _Spell_augmentation:
    def __init__(self, sp_tokenizer, corpus, add_norvig_method = True):
        self._sp_tokenizer = sp_tokenizer
        if self._sp_tokenizer:
            self._augment = _augment_vowel_prob_sp
        else:
            self._augment = _augment_vowel_prob
        self._add_norvig_method = add_norvig_method
        self._corpus = corpus
        self.WORDS = Counter(self._corpus)
        self.N = sum(self.WORDS.values())

    def edit_step(self, word):
        """
        All edits that are one edit away from `word`.
        """
        if self._add_norvig_method:
            splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
            deletes = [L + R[1:] for L, R in splits if R]
            transposes = [
                L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1
            ]
            inserts = [L + c + R for L, R in splits for c in alphabet]
        pseudo = _augment_vowel(word)
        pseudo.extend(self._augment(word, sp_tokenizer = self._sp_tokenizer))
        fuzziness = []
        if len(word):

            # berape -> berapa, mne -> mna
            if word[-1] == 'e':
                inner = word[:-1] + 'a'
                fuzziness.append(inner)
                pseudo.extend(
                    self._augment(inner, sp_tokenizer = self._sp_tokenizer)
                )

            # pikir -> fikir
            if word[0] == 'p':
                inner = 'f' + word[1:]
                fuzziness.append(inner)
                pseudo.extend(
                    self._augment(inner, sp_tokenizer = self._sp_tokenizer)
                )

        if len(word) > 2:
            # bapak -> bapa, mintak -> minta, mntak -> mnta
            if word[-2:] == 'ak':
                fuzziness.append(word[:-1])
                pseudo.extend(
                    self._augment(word[:-1], sp_tokenizer = self._sp_tokenizer)
                )

            # hnto -> hantar, bako -> bkar, sabo -> sabar
            if (
                word[-1] == 'o'
                and word[-3] in vowels
                and word[-2] in consonants
            ):
                inner = word[:-1] + 'ar'
                fuzziness.append(inner)
                pseudo.extend(
                    self._augment(inner, sp_tokenizer = self._sp_tokenizer)
                )

            # antu -> hantu, antar -> hantar
            if word[0] == 'a' and word[1] in consonants:
                inner = 'h' + word
                fuzziness.append(inner)
                pseudo.extend(_augment_vowel(inner))
                pseudo.extend(
                    self._augment(inner, sp_tokenizer = self._sp_tokenizer)
                )

            # ptg -> ptng, dtg -> dtng
            if (
                word[-3] in consonants
                and word[-2] in consonants
                and word[-1] == 'g'
            ):
                inner = word[:-1] + 'ng'
                fuzziness.append(inner)
                pseudo.extend(
                    self._augment(inner, sp_tokenizer = self._sp_tokenizer)
                )

            # igt -> ingt
            if word[1] == 'g' and word[2] in consonants:
                inner = word[0] + 'n' + word[1:]
                fuzziness.append(inner)
                pseudo.extend(
                    self._augment(inner, sp_tokenizer = self._sp_tokenizer)
                )

        if self._add_norvig_method:
            return set(deletes + transposes + inserts + fuzziness + pseudo)
        else:
            return set(fuzziness + pseudo)

    def edits2(self, word):
        """
        All edits that are two edits away from `word`.
        """
        return (e2 for e1 in self.edit_step(word) for e2 in self.edit_step(e1))

    def known(self, words):
        """
        The subset of `words` that appear in the dictionary of WORDS.
        """
        return set(w for w in words if w in self.WORDS)

    def edit_candidates(self, word):
        """
        Generate possible spelling corrections for word.
        """

        ttt = self.known(self.edit_step(word)) or {word}
        ttt = {i for i in ttt if len(i) > 3 and i not in ENGLISH_WORDS}
        ttt = self.known([word]) | ttt
        if not len(ttt):
            ttt = {word}
        return ttt

    def case_of(self, text):
        """
        Return the case-function appropriate for text: upper, lower, title, or just str.
        """

        return (
            str.upper
            if text.isupper()
            else str.lower
            if text.islower()
            else str.title
            if text.istitle()
            else str
        )


class _TransformerCorrector(_Spell_augmentation):
    def __init__(self, model, corpus, sp_tokenizer):
        _Spell_augmentation.__init__(
            self, sp_tokenizer, corpus, add_norvig_method = False
        )
        self._model = model

        import tensorflow as tf

        self._padding = tf.keras.preprocessing.sequence.pad_sequences

    def _correct(self, word, string, index, batch_size = 20):
        possible_states = self.edit_candidates(word)
        replaced_masks = []
        for state in possible_states:
            mask = string[:]
            mask[index] = state
            replaced_masks.append(' '.join(mask))
        ids = [
            generate_ids(mask, self._model._tokenizer)
            for mask in replaced_masks
        ]
        tokens, input_ids, tokens_ids = list(zip(*ids))

        indices, ids = [], []
        for i in range(len(input_ids)):
            indices.extend([i] * len(input_ids[i]))
            ids.extend(input_ids[i])

        masked_padded = self._padding(ids, padding = 'post')
        preds = []
        for i in range(0, len(masked_padded), batch_size):
            index = min(i + batch_size, len(masked_padded))
            batch = masked_padded[i:index]
            preds.append(self._model._log_vectorize(batch))

        preds = np.concatenate(preds, axis = 0)
        indices = np.array(indices)
        scores = []
        for i in range(len(tokens)):
            filter_preds = preds[indices == i]
            total = np.sum(
                [filter_preds[k, k + 1, x] for k, x in enumerate(tokens_ids[i])]
            )
            scores.append(total)

        prob_scores = np.array(scores) / np.sum(scores)
        probs = list(zip(possible_states, prob_scores))
        probs.sort(key = lambda x: x[1])
        return probs[0][0]

    def correct(self, word, string, index = -1, batch_size = 20):
        """
        Correct a word within a text, returning the corrected word.
        """
        if not isinstance(word, str):
            raise ValueError('word must be a string')
        if not isinstance(string, str):
            raise ValueError('string must be a string')
        if not isinstance(batch_size, int):
            raise ValueError('batch_size must be an integer')
        if batch_size < 1:
            raise ValueError('batch_size must be bigger than 0')
        if not isinstance(index, int):
            raise ValueError('index must be an integer')
        string = string.split()
        if word not in string:
            raise ValueError('word not in string after split by spaces')
        if index < 0:
            index = string.index(word)

        if word in ENGLISH_WORDS:
            return word
        if word in MALAY_WORDS:
            return word
        if word in stopword_tatabahasa:
            return word

        if word in rules_normalizer:
            word = rules_normalizer[word]
        else:
            word = self._correct(word, string, index, batch_size = batch_size)
        return word

    def correct_text(self, text, batch_size = 20):
        """
        Correct all the words within a text, returning the corrected text."""

        if not isinstance(text, str):
            raise ValueError('text must be a string')
        if not isinstance(batch_size, int):
            raise ValueError('batch_size must be an integer')
        if batch_size < 1:
            raise ValueError('batch_size must be bigger than 0')

        text = re.sub('[^a-zA-Z]+', ' ', text)
        string = re.sub(r'[ ]+', ' ', text).strip()
        strings = []
        for no, word in enumerate(string.split()):
            if not word[0].isupper():
                word = self.case_of(word)(
                    self.correct(
                        word.lower(), string, no, batch_size = batch_size
                    )
                )
            strings.append(word)

        return ' '.join(strings)

    def correct_word(self, word, string, batch_size = 20):
        """
        Spell-correct word in match, and preserve proper upper/lower/title case.
        """

        return self.case_of(word)(
            self.correct(word.lower(), string, batch_size = batch_size)
        )


class _SpellCorrector(_Spell_augmentation):
    """
    The SpellCorrector extends the functionality of the Peter Norvig's
    spell-corrector in http://norvig.com/spell-correct.html
    And improve it using some algorithms from Normalization of noisy texts in Malaysian online reviews,
    https://www.researchgate.net/publication/287050449_Normalization_of_noisy_texts_in_Malaysian_online_reviews
    Added custom vowels augmentation
    """

    def __init__(self, corpus, sp_tokenizer = None):
        _Spell_augmentation.__init__(self, sp_tokenizer, corpus)

    def tokens(text):
        return REGEX_TOKEN.findall(text.lower())

    def P(self, word):
        """
        Probability of `word`.
        """
        return self.WORDS[word] / self.N

    def most_probable(self, words):
        _known = self.known(words)
        if _known:
            return max(_known, key = self.P)
        else:
            return []

    def correct(self, word, **kwargs):
        """
        Most probable spelling correction for word.
        """
        if not isinstance(word, str):
            raise ValueError('word must be a string')

        if word in ENGLISH_WORDS:
            return word
        if self._corpus.get(word, 0) > 5000:
            return word
        if word in MALAY_WORDS:
            return word
        if word in stopword_tatabahasa:
            return word
        hujung_result = [v for k, v in hujung.items() if word.endswith(k)]
        cp_word = word[:]
        if len(hujung_result):
            hujung_result = max(hujung_result, key = len)
            if len(hujung_result):
                word = word[: -len(hujung_result)]
        permulaan_result = [
            v for k, v in permulaan.items() if word.startswith(k)
        ]
        if len(permulaan_result):
            permulaan_result = max(permulaan_result, key = len)
            if len(permulaan_result):
                word = word[len(permulaan_result) :]
        if len(word):
            if word in rules_normalizer:
                word = rules_normalizer[word]
            elif self._corpus.get(word, 0) > 1000:
                pass
            else:
                candidates1 = self.edit_candidates(word)
                candidates2 = self.edit_candidates(cp_word)
                word1 = max(candidates1, key = self.P)
                word2 = max(candidates2, key = self.P)

                if self.WORDS[word1] > self.WORDS[word2]:
                    word = word1
                    if len(hujung_result) and not word.endswith(hujung_result):
                        word = word + hujung_result
                    if len(permulaan_result) and not word.startswith(
                        permulaan_result
                    ):
                        word = permulaan_result + word
                else:
                    word = word2

        else:
            if len(hujung_result) and not word.endswith(hujung_result):
                word = word + hujung_result
            if len(permulaan_result) and not word.startswith(permulaan_result):
                word = permulaan_result + word

        return word

    def correct_text(self, text):
        """
        Correct all the words within a text, returning the corrected text."""

        if not isinstance(text, str):
            raise ValueError('text must be a string')

        return re.sub('[a-zA-Z]+', self.correct_match, text)

    def correct_match(self, match):
        """
        Spell-correct word in match, and preserve proper upper/lower/title case.
        """

        word = match.group()
        if word[0].isupper():
            return word
        return self.case_of(word)(self.correct(word.lower()))

    def correct_word(self, word):
        """
        Spell-correct word in match, and preserve proper upper/lower/title case.
        """

        return self.case_of(word)(self.correct(word.lower()))

    def elong_normalized_candidates(self, word, acc = None):
        if acc is None:
            acc = []
        candidates = [w for w in set(word) if word.count(w) > 1]
        for c in candidates:
            _w = word.replace(c + c, c)
            if _w in acc:
                continue
            acc.append(_w)
            self.elong_normalized_candidates(_w, acc)
        return acc + [word]

    def best_elong_candidate(self, word):
        candidates = self.elong_normalized_candidates(word)
        best = self.most_probable(candidates)
        return best or word

    def normalize_elongated(self, word):
        return self.case_of(word)(self.best_elong_candidate(word.lower()))


class _SymspellCorrector:
    """
    The SymspellCorrector extends the functionality of symspeller, https://github.com/mammothb/symspellpy
    And improve it using some algorithms from Normalization of noisy texts in Malaysian online reviews,
    https://www.researchgate.net/publication/287050449_Normalization_of_noisy_texts_in_Malaysian_online_reviews
    Added custom vowels augmentation
    """

    def __init__(self, model, verbosity, corpus, k = 10):
        self._model = model
        self._verbosity = verbosity
        self._corpus = corpus
        self.k = k

    def predict(self, word):
        max_edit_distance_lookup = 2
        suggestion_verbosity = self._verbosity
        suggestions = self._model.lookup(
            word, suggestion_verbosity, max_edit_distance_lookup
        )[: self.k]
        return suggestions

    def edit_step(self, word):
        result = list(_augment_vowel_alternate(word))

        if len(word):

            # berape -> berapa, mne -> mna
            if word[-1] == 'e':
                inner = word[:-1] + 'a'
                result.extend(list(_augment_vowel_alternate(inner)))

            # pikir -> fikir
            if word[0] == 'p':
                inner = 'f' + word[1:]
                result.extend(list(_augment_vowel_alternate(inner)))

        if len(word) > 2:
            # bapak -> bapa, mintak -> minta, mntak -> mnta
            if word[-2:] == 'ak':
                fuzziness.append(word[:-1])
                result.extend(list(_augment_vowel_alternate(word[:-1])))

            # hnto -> hantar, bako -> bkar, sabo -> sabar
            if (
                word[-1] == 'o'
                and word[-3] in vowels
                and word[-2] in consonants
            ):
                inner = word[:-1] + 'ar'
                result.extend(list(_augment_vowel_alternate(inner)))

            # antu -> hantu, antar -> hantar
            if word[0] == 'a' and word[1] in consonants:
                inner = 'h' + word
                result.extend(list(_augment_vowel_alternate(inner)))

            # ptg -> ptng, dtg -> dtng
            if (
                word[-3] in consonants
                and word[-2] in consonants
                and word[-1] == 'g'
            ):
                inner = word[:-1] + 'ng'
                result.extend(list(_augment_vowel_alternate(inner)))

            # igt -> ingt
            if word[1] == 'g' and word[2] in consonants:
                inner = word[0] + 'n' + word[1:]
                result.extend(list(_augment_vowel_alternate(inner)))

        words = {}
        for r in result:
            suggestions = self.predict(r)
            for s in suggestions:
                words[s.term] = words.get(s.term, 0) + (
                    s.count / (s.distance + 1)
                )

        return words

    def edit_candidates(self, word):
        ttt = self.edit_step(word) or {word: 10}
        ttt = {
            k: v
            for k, v in ttt.items()
            if len(k) > 3 and k not in ENGLISH_WORDS
        }
        ttt[word] = ttt.get(word, 0) + 10
        if not len(ttt):
            ttt = {word: 10}
        return ttt

    def correct(self, word, **kwargs):
        """
        Most probable spelling correction for word.
        """
        if not isinstance(word, str):
            raise ValueError('word must be a string')

        if word in ENGLISH_WORDS:
            return word
        if self._corpus.get(word, 0) > 5000:
            return word
        if word in MALAY_WORDS:
            return word
        if word in stopword_tatabahasa:
            return word

        cp_word = word[:]
        hujung_result = [v for k, v in hujung.items() if word.endswith(k)]
        if len(hujung_result):
            hujung_result = max(hujung_result, key = len)
            if len(hujung_result):
                word = word[: -len(hujung_result)]
        permulaan_result = [
            v for k, v in permulaan.items() if word.startswith(k)
        ]
        if len(permulaan_result):
            permulaan_result = max(permulaan_result, key = len)
            if len(permulaan_result):
                word = word[len(permulaan_result) :]
        if len(word):
            if word in rules_normalizer:
                word = rules_normalizer[word]
            else:
                candidates1 = self.edit_candidates(word)
                candidates2 = self.edit_candidates(cp_word)
                word1 = max(candidates1, key = candidates1.get)
                word2 = max(candidates2, key = candidates2.get)

                if candidates1[word1] > candidates2[word2]:
                    word = word1
                    if len(hujung_result) and not word.endswith(hujung_result):
                        word = word + hujung_result
                    if len(permulaan_result) and not word.startswith(
                        permulaan_result
                    ):
                        word = permulaan_result + word
                else:
                    word = word2

        else:
            if len(hujung_result) and not word.endswith(hujung_result):
                word = word + hujung_result
            if len(permulaan_result) and not word.startswith(permulaan_result):
                word = permulaan_result + word
        return word

    def correct_text(self, text):
        """
        Correct all the words within a text, returning the corrected text."""

        if not isinstance(text, str):
            raise ValueError('text must be a string')

        return re.sub('[a-zA-Z]+', self.correct_match, text)

    def correct_match(self, match):
        """
        Spell-correct word in match, and preserve proper upper/lower/title case.
        """

        word = match.group()
        if word[0].isupper():
            return word
        return self.case_of(word)(self.correct(word.lower()))

    def case_of(self, text):
        """
        Return the case-function appropriate for text: upper, lower, title, or just str.
        """

        return (
            str.upper
            if text.isupper()
            else str.lower
            if text.islower()
            else str.title
            if text.istitle()
            else str
        )


def probability(sentence_piece = False, validate = True):
    """
    Train a Probability Spell Corrector.

    Parameters
    ----------
    sentence_piece: bool, optional (default=False)
        if True, reduce possible augmentation states using sentence piece.
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    _SpellCorrector: malaya.spell._SpellCorrector class
    """
    if not isinstance(sentence_piece, bool):
        raise ValueError('sentence_piece must be a boolean')

    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    if validate:
        check_file(PATH_NGRAM[1], S3_PATH_NGRAM[1])
    else:
        if not check_available(PATH_NGRAM[1]):
            raise Exception(
                'preprocessing is not available, please `validate = True`'
            )

    tokenizer = None

    if sentence_piece:
        if validate:
            check_file(
                PATH_NGRAM['sentencepiece'], S3_PATH_NGRAM['sentencepiece']
            )
        else:
            if not check_available(PATH_NGRAM[1]):
                raise Exception(
                    'sentence piece is not available, please `validate = True`'
                )

        vocab = PATH_NGRAM['sentencepiece']['vocab']
        vocab_model = PATH_NGRAM['sentencepiece']['model']
        tokenizer = _load_sentencepiece(vocab, vocab_model)

    with open(PATH_NGRAM[1]['model']) as fopen:
        corpus = json.load(fopen)
    return _SpellCorrector(corpus, tokenizer)


def symspell(
    validate = True,
    max_edit_distance_dictionary = 2,
    prefix_length = 7,
    term_index = 0,
    count_index = 1,
    top_k = 10,
):
    """
    Train a symspell Spell Corrector.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    _SpellCorrector: malaya.spell._SymspellCorrector class
    """
    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')
    if not isinstance(max_edit_distance_dictionary, int):
        raise ValueError('max_edit_distance_dictionary must be an integer')
    if not isinstance(prefix_length, int):
        raise ValueError('prefix_length must be an integer')
    if not isinstance(term_index, int):
        raise ValueError('term_index must be an integer')
    if not isinstance(count_index, int):
        raise ValueError('count_index must be an integer')

    if validate:
        check_file(PATH_NGRAM['symspell'], S3_PATH_NGRAM['symspell'])
        check_file(PATH_NGRAM[1], S3_PATH_NGRAM[1])
    else:
        if not check_available(PATH_NGRAM['symspell']):
            raise Exception(
                'preprocessing is not available, please `validate = True`'
            )
        if not check_available(PATH_NGRAM[1]):
            raise Exception(
                'preprocessing is not available, please `validate = True`'
            )
    try:
        from symspellpy.symspellpy import SymSpell, Verbosity
    except:
        raise Exception(
            'symspellpy not installed. Please install it and try again.'
        )
    sym_spell = SymSpell(max_edit_distance_dictionary, prefix_length)
    dictionary_path = PATH_NGRAM['symspell']['model']
    sym_spell.load_dictionary(dictionary_path, term_index, count_index)
    with open(PATH_NGRAM[1]['model']) as fopen:
        corpus = json.load(fopen)
    return _SymspellCorrector(sym_spell, Verbosity.ALL, corpus, k = top_k)


def transformer(model, sentence_piece = False, validate = True):
    """
    Load a Transformer Spell Corrector. Right now only supported BERT and ALBERT.

    Parameters
    ----------
    validate: bool, optional (default=True)
        if True, malaya will check model availability and download if not available.

    Returns
    -------
    _TransformerCorrector: malaya.spell._TransformerCorrector class
    """
    if not hasattr(model, '_log_vectorize'):
        raise ValueError('model must has `_log_vectorize` method')

    if not isinstance(sentence_piece, bool):
        raise ValueError('sentence_piece must be a boolean')

    if not isinstance(validate, bool):
        raise ValueError('validate must be a boolean')

    if validate:
        check_file(PATH_NGRAM[1], S3_PATH_NGRAM[1])
    else:
        if not check_available(PATH_NGRAM[1]):
            raise Exception(
                'preprocessing is not available, please `validate = True`'
            )

    tokenizer = None

    if sentence_piece:
        if validate:
            check_file(
                PATH_NGRAM['sentencepiece'], S3_PATH_NGRAM['sentencepiece']
            )
        else:
            if not check_available(PATH_NGRAM[1]):
                raise Exception(
                    'sentence piece is not available, please `validate = True`'
                )

        vocab = PATH_NGRAM['sentencepiece']['vocab']
        vocab_model = PATH_NGRAM['sentencepiece']['model']
        tokenizer = _load_sentencepiece(vocab, vocab_model)

    with open(PATH_NGRAM[1]['model']) as fopen:
        corpus = json.load(fopen)
    return _TransformerCorrector(model, corpus, tokenizer)



from .. import home

MALAY_TEXT = home + '/dictionary/malay-text.txt'
MALAY_TEXT_200K = home + '/dictionary-200k/malay-text.txt'

PATH_NGRAM = {
    1: {
        'model': home + '/preprocessing/ngram1/bm_1grams.json',
        'version': 'v28',
    },
    2: {
        'model': home + '/preprocessing/ngram2/bm_2grams.json',
        'version': 'v23',
    },
    'symspell': {
        'model': home + '/preprocessing/symspell/bm_1grams.txt',
        'version': 'v28',
    },
    'sentencepiece': {
        'vocab': home + '/bert/sp10m.cased.v4.vocab',
        'model': home + '/bert/sp10m.cased.v4.model',
        'version': 'v31',
    },
}

S3_PATH_NGRAM = {
    1: {'model': 'v27/preprocessing/bm_1grams.json'},
    2: {'model': 'v23/preprocessing/bm_2grams.json'},
    'symspell': {'model': 'v27/preprocessing/bm_1grams.txt'},
    'sentencepiece': {
        'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
        'model': 'bert-bahasa/sp10m.cased.v4.model',
    },
}

PATH_PREPROCESSING = {
    1: {
        'model': home + '/preprocessing/count1/1counts_1grams.json',
        'version': 'v23',
    },
    2: {
        'model': home + '/preprocessing/count2/counts_2grams.json',
        'version': 'v23',
    },
    'english-malay': {
        'model': home + '/preprocessing/english-malay/english-malay-200k.json',
        'version': 'v23',
    },
}

S3_PATH_PREPROCESSING = {
    1: {'model': 'v23/preprocessing/counts_1grams.json'},
    2: {'model': 'v23/preprocessing/counts_2grams.json'},
    'english-malay': {'model': 'v23/preprocessing/english-malay-200k.json'},
}

PATH_STEM = {
    'lstm': {
        'model': home + '/stem/lstm/lstm-stem.pb',
        'setting': home + '/stem/lstm/lstm-stem.json',
        'version': 'v15',
    },
    'bahdanau': {
        'model': home + '/stem/bahdanau/bahdanau-stem.pb',
        'setting': home + '/stem/bahdanau/bahdanau-stem.json',
        'version': 'v15',
    },
    'luong': {
        'model': home + '/stem/luong/luong-stem.pb',
        'setting': home + '/stem/luong/luong-stem.json',
        'version': 'v15',
    },
}

S3_PATH_STEM = {
    'lstm': {
        'model': 'v15/stem/lstm-stem.pb',
        'setting': 'v15/stem/lstm-stem.json',
    },
    'bahdanau': {
        'model': 'v15/stem/bahdanau-stem.pb',
        'setting': 'v15/stem/bahdanau-stem.json',
    },
    'luong': {
        'model': 'v15/stem/luong-stem.pb',
        'setting': 'v15/stem/luong-stem.json',
    },
}

PATH_SUMMARIZE = {
    'news': {
        'model': home + '/summarize/summary-news.pb',
        'setting': home + '/summarize/summary-news.json',
        'version': 'v13',
    },
    'wiki': {
        'model': home + '/summarize/summary-wiki.pb',
        'setting': home + '/summarize/summary-wiki.json',
        'version': 'v13',
    },
}

S3_PATH_SUMMARIZE = {
    'news': {
        'model': 'v13/summarize/summary-news.pb',
        'setting': 'v13/summarize/summary-news.json',
    },
    'wiki': {
        'model': 'v13/summarize/summary-wiki.pb',
        'setting': 'v13/summarize/summary-wiki.json',
    },
}

PATH_TOXIC = {
    'multinomial': {
        'model': home + '/toxicity/multinomial/multinomial-toxicity.pkl',
        'vector': home + '/toxicity/multinomial/multinomial-toxicity-tfidf.pkl',
        'version': 'v30',
    },
    'bert': {
        'base': {
            'model': home + '/toxicity/bert/base/bert-toxicity.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
        'small': {
            'model': home + '/toxicity/bert/small/bert-toxicity.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
    },
    'albert': {
        'base': {
            'model': home + '/toxicity/albert/base/albert-toxicity.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/toxicity/xlnet/base/albert-toxicity.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}

S3_PATH_TOXIC = {
    'multinomial': {
        'model': 'v30/toxicity/multinomial-toxicity.pkl',
        'vector': 'v30/toxicity/multinomial-toxicity-tfidf.pkl',
    },
    'bert': {
        'base': {
            'model': 'v30/toxicity/bert-base-toxicity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
        'small': {
            'model': 'v30/toxicity/bert-small-toxicity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
    },
    'albert': {
        'base': {
            'model': 'v30/toxicity/albert-base-toxicity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/toxicity/xlnet-base-toxicity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_POS = {
    'bert': {
        'base': {
            'model': home + '/pos/bert/base/bert-pos.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'setting': home + '/pos/dictionary-pos.json',
            'version': 'v30',
        },
        'small': {
            'model': home + '/pos/bert/small/bert-pos.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'setting': home + '/pos/dictionary-pos.json',
            'version': 'v30',
        },
    },
    'xlnet': {
        'base': {
            'model': home + '/pos/xlnet/base/xlnet-pos.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'setting': home + '/pos/dictionary-pos.json',
            'version': 'v30',
        }
    },
    'albert': {
        'base': {
            'model': home + '/pos/albert/base/xlnet-pos.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'setting': home + '/pos/dictionary-pos.json',
            'version': 'v30',
        }
    },
}

S3_PATH_POS = {
    'bert': {
        'base': {
            'model': 'v30/pos/bert-base-pos.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
            'setting': 'bert-bahasa/dictionary-pos.json',
        },
        'small': {
            'model': 'v30/pos/bert-small-pos.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
            'setting': 'bert-bahasa/dictionary-pos.json',
        },
    },
    'xlnet': {
        'base': {
            'model': 'v30/pos/xlnet-base-pos.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
            'setting': 'bert-bahasa/dictionary-pos.json',
        }
    },
    'albert': {
        'base': {
            'model': 'v30/pos/albert-base-pos.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
            'setting': 'bert-bahasa/dictionary-pos.json',
        }
    },
}


PATH_LANG_DETECTION = {
    'multinomial': {
        'model': home
        + '/language-detection/multinomial/multinomial-language-detection.pkl',
        'vector': home
        + '/language-detection/multinomial/vectorizer-language-detection.pkl',
        'version': 'v10.2',
    },
    'sgd': {
        'model': home + '/language-detection/sgd/sgd-language-detection.pkl',
        'vector': home
        + '/language-detection/multinomial/vectorizer-language-detection.pkl',
        'version': 'v10.2',
    },
    'deep': {
        'model': home
        + '/language-detection/deep/model.ckpt.data-00000-of-00001',
        'index': home + '/language-detection/deep/model.ckpt.index',
        'meta': home + '/language-detection/deep/model.ckpt.meta',
        'vector': home
        + '/language-detection/multinomial/vectorizer-language-detection.pkl',
        'version': 'v10.2',
    },
}

S3_PATH_LANG_DETECTION = {
    'multinomial': {
        'model': 'v10/language-detection/multinomial-language-detection.pkl',
        'vector': 'v10/language-detection/language-detection-vectorizer.pkl',
    },
    'sgd': {
        'model': 'v10/language-detection/sgd-language-detection.pkl',
        'vector': 'v10/language-detection/language-detection-vectorizer.pkl',
    },
    'deep': {
        'model': 'v10/language-detection/model.ckpt.data-00000-of-00001',
        'index': 'v10/language-detection/model.ckpt.index',
        'meta': 'v10/language-detection/model.ckpt.meta',
        'vector': 'v10/language-detection/language-detection-vectorizer.pkl',
    },
}

PATH_ENTITIES = {
    'bert': {
        'base': {
            'model': home + '/entity/bert/base/bert-entity.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'setting': home + '/entity/dictionary-entities.json',
            'version': 'v30',
        },
        'small': {
            'model': home + '/entity/bert/small/bert-entity.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'setting': home + '/entity/dictionary-entities.json',
            'version': 'v30',
        },
    },
    'xlnet': {
        'base': {
            'model': home + '/entity/xlnet/base/xlnet-entity.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'setting': home + '/entity/dictionary-entities.json',
            'version': 'v30',
        }
    },
    'albert': {
        'base': {
            'model': home + '/entity/albert/base/xlnet-entity.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'setting': home + '/entity/dictionary-entities.json',
            'version': 'v30',
        }
    },
}

S3_PATH_ENTITIES = {
    'bert': {
        'base': {
            'model': 'v30/entity/bert-base-entity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
            'setting': 'bert-bahasa/dictionary-entities.json',
        },
        'small': {
            'model': 'v30/entity/bert-small-entity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
            'setting': 'bert-bahasa/dictionary-entities.json',
        },
    },
    'xlnet': {
        'base': {
            'model': 'v30/entity/xlnet-base-entity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
            'setting': 'bert-bahasa/dictionary-entities.json',
        }
    },
    'albert': {
        'base': {
            'model': 'v30/entity/albert-base-entity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
            'setting': 'bert-bahasa/dictionary-entities.json',
        }
    },
}

PATH_SENTIMENT = {
    'multinomial': {
        'model': home + '/sentiment/multinomial/multinomial-sentiment.pkl',
        'vector': home
        + '/sentiment/multinomial/multinomial-sentiment-tfidf.pkl',
        'version': 'v30',
    },
    'bert': {
        'base': {
            'model': home + '/sentiment/bert/base/bert-sentiment.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
        'small': {
            'model': home + '/sentiment/bert/small/bert-sentiment.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
    },
    'albert': {
        'base': {
            'model': home + '/sentiment/albert/base/albert-sentiment.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/sentiment/xlnet/base/albert-sentiment.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}

S3_PATH_SENTIMENT = {
    'multinomial': {
        'model': 'v30/sentiment/multinomial-sentiment.pkl',
        'vector': 'v30/sentiment/multinomial-sentiment-tfidf.pkl',
    },
    'bert': {
        'base': {
            'model': 'v30/sentiment/bert-base-sentiment.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
        'small': {
            'model': 'v30/sentiment/bert-small-sentiment.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
    },
    'albert': {
        'base': {
            'model': 'v30/sentiment/albert-base-sentiment.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/sentiment/xlnet-base-sentiment.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_SUBJECTIVE = {
    'multinomial': {
        'model': home + '/subjective/multinomial/multinomial-subjective.pkl',
        'vector': home
        + '/subjective/multinomial/multinomial-subjective-tfidf.pkl',
        'version': 'v30',
    },
    'bert': {
        'base': {
            'model': home + '/subjective/bert/base/bert-subjective.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
        'small': {
            'model': home + '/subjective/bert/small/bert-subjective.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
    },
    'albert': {
        'base': {
            'model': home + '/subjective/albert/base/albert-subjective.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/subjective/xlnet/base/albert-subjective.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}

S3_PATH_SUBJECTIVE = {
    'multinomial': {
        'model': 'v10/subjective/multinomial-subjective.pkl',
        'vector': 'v10/subjective/multinomial-subjective-tfidf.pkl',
    },
    'bert': {
        'base': {
            'model': 'v30/subjective/bert-base-subjective.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
        'small': {
            'model': 'v30/subjective/bert-small-subjective.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
    },
    'albert': {
        'base': {
            'model': 'v30/subjective/albert-base-subjective.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/subjective/xlnet-base-subjective.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_EMOTION = {
    'multinomial': {
        'model': home + '/emotion/multinomial/multinomial-emotion.pkl',
        'vector': home + '/emotion/multinomial/multinomial-emotion-tfidf.pkl',
        'version': 'v30',
    },
    'bert': {
        'base': {
            'model': home + '/emotion/bert/base/bert-emotion.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
        'small': {
            'model': home + '/emotion/bert/small/bert-emotion.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        },
    },
    'albert': {
        'base': {
            'model': home + '/emotion/albert/base/albert-emotion.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/emotion/xlnet/base/albert-emotion.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}

S3_PATH_EMOTION = {
    'multinomial': {
        'model': 'v30/emotion/multinomial-emotion.pkl',
        'vector': 'v30/emotion/multinomial-emotion-tfidf.pkl',
    },
    'bert': {
        'base': {
            'model': 'v30/emotion/bert-base-emotion.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
        'small': {
            'model': 'v30/emotion/bert-small-emotion.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        },
    },
    'albert': {
        'base': {
            'model': 'v30/emotion/albert-base-emotion.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/emotion/xlnet-base-emotion.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_DEPEND = {
    'bert': {
        'base': {
            'model': home + '/dependency/bert/base/bert-dependency.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        }
    },
    'albert': {
        'base': {
            'model': home + '/dependency/albert/base/albert-dependency.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/dependency/xlnet/base/xlnet-dependency.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}

S3_PATH_DEPEND = {
    'bert': {
        'base': {
            'model': 'v30/dependency/bert-base-dependency.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        }
    },
    'albert': {
        'base': {
            'model': 'v30/dependency/albert-base-dependency.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/dependency/xlnet-base-dependency.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_RELEVANCY = {
    'bert': {
        'base': {
            'model': home + '/relevancy/bert/base/bert-relevancy.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        }
    },
    'albert': {
        'base': {
            'model': home + '/relevancy/albert/base/albert-relevancy.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/relevancy/xlnet/base/albert-relevancy.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}
S3_PATH_RELEVANCY = {
    'bert': {
        'base': {
            'model': 'v30/relevancy/bert-base-relevancy.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        }
    },
    'albert': {
        'base': {
            'model': 'v30/relevancy/albert-base-relevancy.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/relevancy/xlnet-base-relevancy.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_SIMILARITY = {
    'bert': {
        'base': {
            'model': home + '/similarity/bert/base/bert-similarity.pb',
            'vocab': home + '/bert/sp10m.cased.v4.vocab',
            'tokenizer': home + '/bert/sp10m.cased.v4.model',
            'version': 'v30',
        }
    },
    'albert': {
        'base': {
            'model': home + '/similarity/albert/base/albert-similarity.pb',
            'vocab': home + '/albert/sp10m.cased.v6.vocab',
            'tokenizer': home + '/albert/sp10m.cased.v6.model',
            'version': 'v30',
        }
    },
    'xlnet': {
        'base': {
            'model': home + '/similarity/xlnet/base/albert-similarity.pb',
            'vocab': home + '/xlnet/sp10m.cased.v5.vocab',
            'tokenizer': home + '/xlnet/sp10m.cased.v5.model',
            'version': 'v30',
        }
    },
}

S3_PATH_SIMILARITY = {
    'bert': {
        'base': {
            'model': 'v30/similarity/bert-base-similarity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v4.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v4.model',
        }
    },
    'albert': {
        'base': {
            'model': 'v30/similarity/albert-base-similarity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v6.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v6.model',
        }
    },
    'xlnet': {
        'base': {
            'model': 'v30/similarity/xlnet-base-similarity.pb',
            'vocab': 'bert-bahasa/sp10m.cased.v5.vocab',
            'tokenizer': 'bert-bahasa/sp10m.cased.v5.model',
        }
    },
}

PATH_BERT = {
    'base': {
        'path': home + '/bert-model/base',
        'directory': home + '/bert-model/base/bert-base-v2/',
        'model': {
            'model': home + '/bert-model/base/bert-bahasa-base.tar.gz',
            'version': 'v31',
        },
    },
    'small': {
        'path': home + '/bert-model/small',
        'directory': home + '/bert-model/small/bert-small-v2/',
        'model': {
            'model': home + '/bert-model/small/bert-bahasa-small.tar.gz',
            'version': 'v31',
        },
    },
}

S3_PATH_BERT = {
    'base': {'model': 'bert-bahasa/bert-base-2-12-2019.tar.gz'},
    'small': {'model': 'bert-bahasa/bert-small-4-12-2019.tar.gz'},
}

PATH_XLNET = {
    'base': {
        'path': home + '/xlnet-model/base',
        'directory': home + '/xlnet-model/base/xlnet-base/',
        'model': {
            'model': home + '/xlnet-model/base/xlnet-base.tar.gz',
            'version': 'v30',
        },
    }
}

S3_PATH_XLNET = {'base': {'model': 'bert-bahasa/xlnet-base-30-9-2019.tar.gz'}}

PATH_ALBERT = {
    'base': {
        'path': home + '/albert-model/base',
        'directory': home + '/albert-model/base/albert-base/',
        'model': {
            'model': home + '/albert-model/base/albert-base.tar.gz',
            'version': 'v31',
        },
    },
    'large': {
        'path': home + '/albert-model/large',
        'directory': home + '/albert-model/large/albert-large/',
        'model': {
            'model': home + '/albert-model/base/albert-large.tar.gz',
            'version': 'v31',
        },
    },
}

S3_PATH_ALBERT = {
    'base': {'model': 'bert-bahasa/albert-base-30-11-2019.tar.gz'},
    'large': {'model': 'bert-bahasa/albert-large-30-11-2019.tar.gz'},
}

PATH_ALXLNET = {
    'base': {
        'path': home + '/alxlnet-model/base',
        'directory': home + '/alxlnet-model/base/alxlnet-base/',
        'model': {
            'model': home + '/alxlnet-model/base/alxlnet-base.tar.gz',
            'version': 'v31',
        },
    }
}

S3_PATH_ALXLNET = {
    'base': {'model': 'bert-bahasa/alxlnet-base-6-11-2019.tar.gz'}
}



import malaya
string = 'KUALA LUMPUR: Sempena sambutan Aidilfitri minggu depan, Perdana Menteri Tun Dr Mahathir Mohamad dan Menteri Pengangkutan Anthony Loke Siew Fook'

malaya.get_available_pos_models()

def test_pos():
    assert len(malaya.naive_pos(string))

def test_multinomial_pos():
    assert len(malaya.multinomial_pos(string))

def test_xgb_pos():
    assert len(malaya.xgb_pos(string))

def test_deep_pos_char():
    model = malaya.deep_pos('char')
    assert len(model.predict(string))

def test_deep_pos_word():
    model = malaya.deep_pos('word')
    assert len(model.predict(string))

def test_deep_pos_concat():
    model = malaya.deep_pos('concat')
    assert len(model.predict(string))



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions and classes related to optimization (weight updates)."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.training.optimizer import Optimizer
from tensorflow.python.ops import state_ops
from tensorflow.python.ops import resource_variable_ops


def create_optimizer(
    loss, init_lr, num_train_steps, num_warmup_steps, fp16 = False
):
    """Creates an optimizer training op."""
    global_step = tf.train.get_or_create_global_step()

    learning_rate = tf.constant(value = init_lr, shape = [], dtype = tf.float32)

    # Implements linear decay of the learning rate.
    learning_rate = tf.train.polynomial_decay(
        learning_rate,
        global_step,
        num_train_steps,
        end_learning_rate = 0.0,
        power = 1.0,
        cycle = False,
    )

    # Implements linear warmup. I.e., if global_step < num_warmup_steps, the
    # learning rate will be `global_step/num_warmup_steps * init_lr`.
    if num_warmup_steps:
        global_steps_int = tf.cast(global_step, tf.int32)
        warmup_steps_int = tf.constant(num_warmup_steps, dtype = tf.int32)

        global_steps_float = tf.cast(global_steps_int, tf.float32)
        warmup_steps_float = tf.cast(warmup_steps_int, tf.float32)

        warmup_percent_done = global_steps_float / warmup_steps_float
        warmup_learning_rate = init_lr * warmup_percent_done

        is_warmup = tf.cast(global_steps_int < warmup_steps_int, tf.float32)
        learning_rate = (
            1.0 - is_warmup
        ) * learning_rate + is_warmup * warmup_learning_rate

    # It is recommended that you use this optimizer for fine tuning, since this
    # is how the model was trained (note that the Adam m/v variables are NOT
    # loaded from init_checkpoint.)
    optimizer = AdamWeightDecayOptimizer(
        learning_rate = learning_rate,
        weight_decay_rate = 0.01,
        beta_1 = 0.9,
        beta_2 = 0.999,
        epsilon = 1e-6,
        exclude_from_weight_decay = ['LayerNorm', 'layer_norm', 'bias'],
    )

    # REF: https://github.com/tensorflow/tensorflow/issues/25080
    # if fp16:
    #     loss_scale_manager = tf.contrib.mixed_precision.ExponentialUpdateLossScaleManager(
    #         init_loss_scale=2 ** 32,
    #         incr_every_n_steps=1000,
    #         decr_every_n_nan_or_inf=2,
    #         decr_ratio=0.5)
    #     optimizer = tf.contrib.mixed_precision.LossScaleOptimizer(optimizer, loss_scale_manager)

    tvars = tf.trainable_variables()
    gvs = optimizer.compute_gradients(loss, tvars)
    gvs = [(g, v) for g, v in gvs if g is not None]
    grads, tvars = list(zip(*gvs))
    if fp16:
        all_finite = tf.reduce_all(
            [tf.reduce_all(tf.is_finite(g)) for g in grads]
        )
    else:
        all_finite = tf.constant(True, dtype = tf.bool)

    # This is how the model was pre-trained.
    (grads, _) = tf.clip_by_global_norm(
        grads,
        clip_norm = 1.0,
        use_norm = tf.cond(
            all_finite, lambda: tf.global_norm(grads), lambda: tf.constant(1.0)
        ),
    )

    train_op = optimizer.apply_gradients(
        zip(grads, tvars), global_step = global_step
    )

    # Normally the global step update is done inside of `apply_gradients`.
    # However, `AdamWeightDecayOptimizer` doesn't do this. But if you use
    # a different optimizer, you should probably take this line out.
    new_global_step = tf.cond(
        all_finite, lambda: global_step + 1, lambda: global_step
    )
    new_global_step = tf.identity(new_global_step, name = 'update_step')
    train_op = tf.group(train_op, [global_step.assign(new_global_step)])
    return train_op


class AdamWeightDecayOptimizer(Optimizer):
    """A basic Adam optimizer that includes "correct" L2 weight decay."""

    def __init__(
        self,
        learning_rate,
        weight_decay_rate = 0.0,
        beta_1 = 0.9,
        beta_2 = 0.999,
        epsilon = 1e-6,
        exclude_from_weight_decay = None,
        name = 'AdamWeightDecayOptimizer',
    ):
        """Constructs a AdamWeightDecayOptimizer."""
        super(AdamWeightDecayOptimizer, self).__init__(False, name)

        self.learning_rate = tf.identity(learning_rate, name = 'learning_rate')
        self.weight_decay_rate = weight_decay_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.epsilon = epsilon
        self.exclude_from_weight_decay = exclude_from_weight_decay

    def _prepare(self):
        self.learning_rate_t = ops.convert_to_tensor(
            self.learning_rate, name = 'learning_rate'
        )
        self.weight_decay_rate_t = ops.convert_to_tensor(
            self.weight_decay_rate, name = 'weight_decay_rate'
        )
        self.beta_1_t = ops.convert_to_tensor(self.beta_1, name = 'beta_1')
        self.beta_2_t = ops.convert_to_tensor(self.beta_2, name = 'beta_2')
        self.epsilon_t = ops.convert_to_tensor(self.epsilon, name = 'epsilon')

    def _create_slots(self, var_list):
        for v in var_list:
            self._zeros_slot(v, 'm', self._name)
            self._zeros_slot(v, 'v', self._name)

    def _apply_dense(self, grad, var):
        learning_rate_t = math_ops.cast(
            self.learning_rate_t, var.dtype.base_dtype
        )
        beta_1_t = math_ops.cast(self.beta_1_t, var.dtype.base_dtype)
        beta_2_t = math_ops.cast(self.beta_2_t, var.dtype.base_dtype)
        epsilon_t = math_ops.cast(self.epsilon_t, var.dtype.base_dtype)
        weight_decay_rate_t = math_ops.cast(
            self.weight_decay_rate_t, var.dtype.base_dtype
        )

        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')

        # Standard Adam update.
        next_m = tf.multiply(beta_1_t, m) + tf.multiply(1.0 - beta_1_t, grad)
        next_v = tf.multiply(beta_2_t, v) + tf.multiply(
            1.0 - beta_2_t, tf.square(grad)
        )

        update = next_m / (tf.sqrt(next_v) + epsilon_t)

        if self._do_use_weight_decay(var.name):
            update += weight_decay_rate_t * var

        update_with_lr = learning_rate_t * update

        next_param = var - update_with_lr

        return control_flow_ops.group(
            *[var.assign(next_param), m.assign(next_m), v.assign(next_v)]
        )

    def _resource_apply_dense(self, grad, var):
        learning_rate_t = math_ops.cast(
            self.learning_rate_t, var.dtype.base_dtype
        )
        beta_1_t = math_ops.cast(self.beta_1_t, var.dtype.base_dtype)
        beta_2_t = math_ops.cast(self.beta_2_t, var.dtype.base_dtype)
        epsilon_t = math_ops.cast(self.epsilon_t, var.dtype.base_dtype)
        weight_decay_rate_t = math_ops.cast(
            self.weight_decay_rate_t, var.dtype.base_dtype
        )

        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')

        # Standard Adam update.
        next_m = tf.multiply(beta_1_t, m) + tf.multiply(1.0 - beta_1_t, grad)
        next_v = tf.multiply(beta_2_t, v) + tf.multiply(
            1.0 - beta_2_t, tf.square(grad)
        )

        update = next_m / (tf.sqrt(next_v) + epsilon_t)

        if self._do_use_weight_decay(var.name):
            update += weight_decay_rate_t * var

        update_with_lr = learning_rate_t * update

        next_param = var - update_with_lr

        return control_flow_ops.group(
            *[var.assign(next_param), m.assign(next_m), v.assign(next_v)]
        )

    def _apply_sparse_shared(self, grad, var, indices, scatter_add):
        learning_rate_t = math_ops.cast(
            self.learning_rate_t, var.dtype.base_dtype
        )
        beta_1_t = math_ops.cast(self.beta_1_t, var.dtype.base_dtype)
        beta_2_t = math_ops.cast(self.beta_2_t, var.dtype.base_dtype)
        epsilon_t = math_ops.cast(self.epsilon_t, var.dtype.base_dtype)
        weight_decay_rate_t = math_ops.cast(
            self.weight_decay_rate_t, var.dtype.base_dtype
        )

        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')

        m_t = state_ops.assign(m, m * beta_1_t, use_locking = self._use_locking)

        m_scaled_g_values = grad * (1 - beta_1_t)
        with ops.control_dependencies([m_t]):
            m_t = scatter_add(m, indices, m_scaled_g_values)

        v_scaled_g_values = (grad * grad) * (1 - beta_2_t)
        v_t = state_ops.assign(v, v * beta_2_t, use_locking = self._use_locking)
        with ops.control_dependencies([v_t]):
            v_t = scatter_add(v, indices, v_scaled_g_values)

        update = m_t / (math_ops.sqrt(v_t) + epsilon_t)

        if self._do_use_weight_decay(var.name):
            update += weight_decay_rate_t * var

        update_with_lr = learning_rate_t * update

        var_update = state_ops.assign_sub(
            var, update_with_lr, use_locking = self._use_locking
        )
        return control_flow_ops.group(*[var_update, m_t, v_t])

    def _apply_sparse(self, grad, var):
        return self._apply_sparse_shared(
            grad.values,
            var,
            grad.indices,
            lambda x, i, v: state_ops.scatter_add(  # pylint: disable=g-long-lambda
                x, i, v, use_locking = self._use_locking
            ),
        )

    def _resource_scatter_add(self, x, i, v):
        with ops.control_dependencies(
            [resource_variable_ops.resource_scatter_add(x.handle, i, v)]
        ):
            return x.value()

    def _resource_apply_sparse(self, grad, var, indices):
        return self._apply_sparse_shared(
            grad, var, indices, self._resource_scatter_add
        )

    def _do_use_weight_decay(self, param_name):
        """Whether to use L2 weight decay for `param_name`."""
        if not self.weight_decay_rate:
            return False
        if self.exclude_from_weight_decay:
            for r in self.exclude_from_weight_decay:
                if re.search(r, param_name) is not None:
                    return False
        return True
